from abc import abstractmethod, ABC
from dataclasses import dataclass
from gzip import GzipFile
from pathlib import Path
from typing import Sequence, NamedTuple, Iterator, Pattern, Iterable
from urllib.parse import quote, urljoin

from bs4 import Tag, BeautifulSoup
from tqdm.auto import tqdm

from archive_query_log.legacy.download.iterable import ArchivedRawSerps
from archive_query_log.legacy.model import ArchivedRawSerp, \
    ArchivedSearchResultSnippet, ResultsParser, InterpretedQueryParser, \
    ArchivedParsedSerp, Service, HighlightedText
from archive_query_log.legacy.util.html import clean_html
from archive_query_log.legacy.util.text import text_io_wrapper


class HtmlResultsParser(ResultsParser, ABC):
    url_pattern: Pattern[str]

    @abstractmethod
    def parse_html(
            self,
            html: Tag,
            timestamp: int,
            serp_url: str,
    ) -> Iterator[ArchivedSearchResultSnippet]:
        ...

    def parse(
            self,
            raw_serp: ArchivedRawSerp,
    ) -> Sequence[ArchivedSearchResultSnippet] | None:
        if self.url_pattern.search(raw_serp.url) is None:
            return None
        html = BeautifulSoup(
            raw_serp.content,
            "html.parser",
            from_encoding=raw_serp.encoding
        )
        results = tuple(self.parse_html(
            html,
            raw_serp.timestamp,
            raw_serp.url,
        ))
        return results if len(results) > 0 else None


@dataclass(frozen=True)
class HtmlSelectorResultsParser(HtmlResultsParser):
    url_pattern: Pattern[str]
    results_selector: str
    url_selector: str
    url_attribute: str
    title_selector: str
    snippet_selector: str | None

    def parse_html(
            self,
            html: Tag,
            timestamp: int,
            serp_url: str,
    ) -> Iterator[ArchivedSearchResultSnippet]:
        for index, result in enumerate(html.select(self.results_selector)):
            url_tag: Tag | None
            if self.url_selector == ":--self":
                url_tag = result
            else:
                url_tag = result.select_one(self.url_selector)
            if url_tag is None:
                continue
            if self.url_attribute not in url_tag.attrs:
                continue
            url = url_tag.attrs[self.url_attribute]
            if url is None:
                continue
            url = urljoin(serp_url, url)

            title_tag: Tag | None
            if self.title_selector == ":--self":
                title_tag = result
            else:
                title_tag = result.select_one(self.title_selector)
            if title_tag is None:
                continue
            title = HighlightedText(clean_html(title_tag))
            if len(title) == 0:
                continue

            snippet = None
            if self.snippet_selector is not None:
                if self.snippet_selector == ":--self":
                    snippet_tags = [result]
                else:
                    snippet_tags = result.select(self.snippet_selector)
                if snippet_tags is not None and snippet_tags:
                    for snippet_candidate_tag in snippet_tags:
                        snippet_candidate = HighlightedText(
                            clean_html(snippet_candidate_tag)
                        )

                        if (len(snippet_candidate) > 0 and
                                (not snippet or
                                 len(snippet_candidate) > len(snippet))):
                            snippet = snippet_candidate

            yield ArchivedSearchResultSnippet(
                rank=index + 1,
                url=url,
                timestamp=timestamp,
                title=title,
                snippet=snippet,
            )


class HtmlInterpretedQueryParser(InterpretedQueryParser, ABC):
    url_pattern: Pattern[str]

    @abstractmethod
    def parse_html(self, html: Tag) -> str | None:
        ...

    def parse(
            self,
            raw_serp: ArchivedRawSerp,
    ) -> str | None:
        if self.url_pattern.search(raw_serp.url) is None:
            return None
        html = BeautifulSoup(
            raw_serp.content,
            "html.parser",
            from_encoding=raw_serp.encoding
        )
        return self.parse_html(html)


@dataclass(frozen=True)
class HtmlSelectorInterpretedQueryParser(HtmlInterpretedQueryParser):
    url_pattern: Pattern[str]
    query_selector: str
    query_attribute: str
    query_text: bool = False

    def parse_html(self, html: Tag) -> str | None:
        search_field = html.select_one(self.query_selector)
        if search_field is None:
            return None
        if self.query_text:
            interpreted_query = search_field.text
            if interpreted_query is not None and len(interpreted_query) > 0:
                return interpreted_query
        if (self.query_attribute is not None and
                self.query_attribute in search_field.attrs):
            interpreted_query = search_field.attrs[self.query_attribute]
            if interpreted_query is not None and len(interpreted_query) > 0:
                return interpreted_query
        return None


class _CdxPage(NamedTuple):
    input_path: Path
    output_path: Path


@dataclass(frozen=True)
class ArchivedParsedSerpParser:
    results_parsers: Sequence[ResultsParser]
    interpreted_query_parsers: Sequence[InterpretedQueryParser]
    overwrite: bool = False
    verbose: bool = False

    def parse(self, input_path: Path, output_path: Path) -> None:
        if output_path.exists() and not self.overwrite:
            return
        output_path.parent.mkdir(parents=True, exist_ok=True)
        archived_serp_contents: Iterable[ArchivedRawSerp] = (
            ArchivedRawSerps(input_path))
        if self.verbose:
            # noinspection PyTypeChecker
            archived_serp_contents = tqdm(
                archived_serp_contents,
                desc="Parse SERP WARC records",
                unit="record",
            )
        archived_parsed_serps_nullable = (
            self.parse_single(archived_serp_content)
            for archived_serp_content in archived_serp_contents
        )
        archived_parsed_serps = (
            archived_serp
            for archived_serp in archived_parsed_serps_nullable
            if archived_serp is not None
        )
        output_schema = ArchivedParsedSerp.schema()
        with output_path.open("wb") as file, \
                GzipFile(fileobj=file, mode="wb") as gzip_file, \
                text_io_wrapper(gzip_file) as text_file:
            for archived_parsed_serp in archived_parsed_serps:
                text_file.write(output_schema.dumps(archived_parsed_serp))
                text_file.write("\n")

    def parse_single(
            self,
            archived_serp_content: ArchivedRawSerp
    ) -> ArchivedParsedSerp | None:
        results: Sequence[ArchivedSearchResultSnippet] | None = None
        for results_parser in self.results_parsers:
            results = results_parser.parse(archived_serp_content)
            if results is not None:
                break

        interpreted_query: str | None = None
        for interpreted_query_parser in self.interpreted_query_parsers:
            interpreted_query = interpreted_query_parser.parse(
                archived_serp_content)
            if interpreted_query is not None:
                break

        if results is None and interpreted_query is None:
            return None

        return ArchivedParsedSerp(
            url=archived_serp_content.url,
            timestamp=archived_serp_content.timestamp,
            query=archived_serp_content.query,
            page=archived_serp_content.page,
            offset=archived_serp_content.offset,
            interpreted_query=interpreted_query,
            results=results if results is not None else [],
        )

    def _service_pages(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None,
            cdx_page: int | None,
    ) -> Sequence[_CdxPage]:
        """
        List all items that need to be downloaded.
        """
        input_format_path = data_directory / "archived-raw-serps"
        output_format_path = data_directory / "archived-parsed-serps"

        service_path = input_format_path / service.name
        if not service_path.exists():
            return []

        if domain is not None:
            domain_paths = [service_path / domain]
        else:
            domain_paths = [
                path
                for path in service_path.iterdir()
                if path.is_dir()
            ]
            if focused:
                domain_paths = [
                    path
                    for path in domain_paths
                    if any(
                        path.name.endswith(quote(prefix, safe=""))
                        for prefix in service.focused_url_prefixes
                    )
                ]

        if cdx_page is not None:
            if domain is None:
                raise RuntimeError(
                    "Domain must be specified when page is specified.")
            if len(domain_paths) < 1:
                raise RuntimeError(
                    "There must be exactly one domain path.")
            cdx_page_paths = [domain_paths[0] / f"{cdx_page:010}"]
        else:
            cdx_page_paths = [
                path
                for domain_path in domain_paths
                for path in domain_path.iterdir()
                if (
                        path.is_dir() and
                        len(path.name) == 10 and
                        path.name.isdigit()
                )
            ]

        return [
            _CdxPage(
                input_path=cdx_page_path,
                output_path=output_format_path / cdx_page_path.relative_to(
                    input_format_path
                ).with_suffix(".jsonl.gz"),
            )
            for cdx_page_path in cdx_page_paths
        ]

    def parse_service(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None = None,
            cdx_page: int | None = None,
    ):
        pages_list: Sequence[_CdxPage] = self._service_pages(
            data_directory=data_directory,
            focused=focused,
            service=service,
            domain=domain,
            cdx_page=cdx_page,
        )

        if len(pages_list) == 0:
            return

        pages: Iterable[_CdxPage] = pages_list
        if len(pages_list) > 1:
            # noinspection PyTypeChecker
            pages = tqdm(
                pages,
                desc="Parse archived SERP URLs",
                unit="page",
            )

        for page in pages:
            self.parse(page.input_path, page.output_path)



from asyncio import sleep
from pathlib import Path
from random import random
from typing import Iterable, Callable, Mapping

from aiohttp import ClientResponseError
from aiohttp_retry import RetryClient
from asyncio_pool import AioPool
from tqdm.auto import tqdm

from archive_query_log.legacy.model import ArchivedUrl
from archive_query_log.legacy.util.archive_http import archive_http_client


class WebArchiveRawDownloader:

    async def download(
            self,
            archived_urls: Iterable[ArchivedUrl],
            download_directory_path: Path,
            file_name: Callable[[ArchivedUrl], str],
    ) -> Mapping[ArchivedUrl, Path]:
        """
        Download the original HTML of a collection of archived URLs
        from the Internet Archive's Wayback Machine and
        return the paths to the individual downloaded files.

        The downloader will retry requests with a backoff and will continue
        downloading URLs even if some URLs fail.

        :param archived_urls: The archived URLs to download.
        :param download_directory_path: Path to the directory in which the
         downloaded HTML files should be stored.
        :param file_name: Function for specifying each URL's file name
         in the download directory.
        :return: A mapping of the successfully downloaded URLs to their
         corresponding download file paths. Failed downloads will not appear
         in the mapping.
        """
        archived_urls = list(archived_urls)
        progress = tqdm(
            total=len(archived_urls),
            desc="Download archived URLs",
            unit="URL",
        )
        async with archive_http_client(limit=10) as client:
            pool = AioPool(size=100)  # avoid creating too many tasks at once

            async def download_single(archived_url: ArchivedUrl):
                return await self._download_single_progress(
                    archived_url,
                    download_directory_path,
                    file_name,
                    client,
                    progress,
                )

            responses = await pool.map(download_single, archived_urls)
        return {
            archived_url: response
            for archived_url, response in zip(archived_urls, responses)
            if response is not None
        }

    async def download_single(
            self,
            archived_url: ArchivedUrl,
            download_directory_path: Path,
            file_name: Callable[[ArchivedUrl], str],
    ) -> Path | None:
        """
        Download the original HTML of a single archived URL
        from the Internet Archive's Wayback Machine and
        return the path to the downloaded file.

        The downloader will retry requests with a backoff.

        :param archived_url: The archived URL to download.
        :param download_directory_path: Path to the directory in which the
         downloaded HTML file should be stored.
        :param file_name: Function for specifying each URL's file name
         in the download directory.
        :return: The download file path if the request was successful,
         and None otherwise.
        """
        async with archive_http_client(limit=1) as client:
            return await self._download_single(
                archived_url=archived_url,
                download_directory_path=download_directory_path,
                file_name=file_name,
                client=client,
            )

    @staticmethod
    async def _download_single(
            archived_url: ArchivedUrl,
            download_directory_path: Path,
            file_name: Callable[[ArchivedUrl], str],
            client: RetryClient,
    ) -> Path | None:
        file_path = download_directory_path / file_name(archived_url)
        if file_path.exists():
            return file_path
        url = archived_url.raw_archive_url
        await sleep(1.0 * random())  # nosec: B311
        try:
            async with client.get(url) as response:
                response.raise_for_status()
                with file_path.open("wb") as file:
                    async for data, _ in response.content.iter_chunks():
                        file.write(data)
                return file_path
        except ClientResponseError:
            file_path.unlink(missing_ok=True)
            return None
        except BaseException as e:
            file_path.unlink(missing_ok=True)
            raise e

    @staticmethod
    async def _download_single_progress(
            archived_url: ArchivedUrl,
            download_directory_path: Path,
            file_name: Callable[[ArchivedUrl], str],
            client: RetryClient,
            progress: tqdm,
    ) -> Path | None:
        res = await WebArchiveRawDownloader._download_single(
            archived_url=archived_url,
            download_directory_path=download_directory_path,
            file_name=file_name,
            client=client,
        )
        progress.update(1)
        return res



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_jd_yu_diao_jiu_ning_meng_1601078153():
    verify_serp_parsing(
        "https://web.archive.org/web/20200925235553id_/https://search.jd.com/Search?keyword=%E9%A2%84%E8%B0%83%E9%85%92%E6%9F%A0%E6%AA%AC&enc=utf-8",
        "jd",
    )


def test_parse_query_jd_qing_shang_1510505248():
    verify_serp_parsing(
        "https://web.archive.org/web/20171112164728id_/http://search.jd.com/search?keyword=%E6%83%85%E5%95%86&enc=utf-8&qrst=1&rt=1&stop=1&book=y&vt=2&wq=%E6%83%85%E5%95%86&ev=publishers_%E5%90%89%E6%9E%97%E5%87%BA%E7%89%88%E9%9B%86%E5%9B%A2%40packstate_no%40&uc=0",
        "jd",
    )


def test_parse_query_jd_tai_kong_zhen_1508948206():
    verify_serp_parsing(
        "https://web.archive.org/web/20171025161646id_/http://search.jd.com/search?keyword=%E5%A4%AA%E7%A9%BA%E6%9E%95&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&stock=1&cid3=12198",
        "jd",
    )


def test_parse_query_jd_xian_xing_dai_shu_1497236213():
    verify_serp_parsing(
        "https://web.archive.org/web/20170612025653id_/http://search.jd.com/search?keyword=%E7%BA%BF%E6%80%A7%E4%BB%A3%E6%95%B0&enc=utf-8&qrst=1&rt=1&stop=1&spm=2.1.1&vt=2&cid3=3672",
        "jd",
    )


def test_parse_query_jd_iu_vdun35358z_1537907650():
    verify_serp_parsing(
        "https://web.archive.org/web/20180925203410id_/http://search.jd.com/search?keyword=%BC%D2%CD%A5%C2%D2%C2%D7%B5%E7%D3%B0%CD%F8%A1%BA%2B%CE%A2v%D0%C5dun35358%A1%BB%A8z&qr=&qrst=UNEXPAND&et=&rt=1&ev=exprice_0-19%40&uc=0",
        "jd",
    )


def test_parse_query_jd_tuan_dui_guan_li_1429813603():
    verify_serp_parsing(
        "https://web.archive.org/web/20150423182643id_/http://search.jd.com/search?keyword=%E5%9B%A2%E9%98%9F%E7%AE%A1%E7%90%86&enc=utf-8&qrst=1&rt=1&stop=1&book=y&499055734",
        "jd",
    )


def test_parse_query_jd_che_zai_cd_1507664971():
    verify_serp_parsing(
        "https://web.archive.org/web/20171010194931id_/http://search.jd.com/search?keyword=%E8%BD%A6%E8%BD%BDCD&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&wq=%E8%BD%A6%E8%BD%BDCD&ev=exstarlevel_2%40128_2382%40&uc=0",
        "jd",
    )


def test_parse_query_jd_ch_1382528052():
    verify_serp_parsing(
        "https://web.archive.org/web/20131023113412id_/http://search.jd.com/search?keyword=%C4%DA%BF%E3%20%C3%DE&qr=%C4%D0%CA%BF%C4%DA%BF%E3&qrst=expand&et=4&rt=1&sttr=1&area=1&ev=1373_71335%40&uc=0",
        "jd",
    )


def test_parse_query_jd_1368578723():
    verify_serp_parsing(
        "https://web.archive.org/web/20130515004523id_/http://search.jd.com/search?keyword=%B4%B4%D0%C2&qr=&qrst=UNEXPAND&et=&rt=1&area=1&ev=&uc=0&bs=no",
        "jd",
    )


def test_parse_query_jd_guan_yin_liao_1501752634():
    verify_serp_parsing(
        "https://web.archive.org/web/20170803093034id_/http://search.jd.com/search?keyword=%E9%A6%86%E9%A5%AE%E6%96%99&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&wq=%E9%A6%86%E9%A5%AE%E6%96%99&cid2=12202&stock=1&ev=2762_32941%40&uc=0",
        "jd",
    )


def test_parse_query_jd_yang_hong_wei_1429640037():
    verify_serp_parsing(
        "https://web.archive.org/web/20150421181357id_/http://search.jd.com/search?keyword=%E6%9D%A8%E5%AE%8F%E4%BC%9F&ev=&psort=2&book=y",
        "jd",
    )


def test_parse_query_jd_weiwei776699e_1537978473():
    verify_serp_parsing(
        "https://web.archive.org/web/20180926161433id_/http://search.jd.com/search?keyword=%BA%FE%B1%B1%BF%EC%C8%FD%CD%F8%D2%D7%BF%AA%BD%B1%BD%E1%B9%FB%CE%A2%D0%C5weiwei776699%85e&qr=&qrst=UNEXPAND&et=&rt=1&ev=exprice_0-19%40&uc=0",
        "jd",
    )


def test_parse_query_jd_shlrwanh_uq_2179706936_1473350033():
    verify_serp_parsing(
        "https://web.archive.org/web/20160908155353id_/http://search.jd.com/Search?keyword=%D5%E6%D5%FD%B1%ED%D1%DD%CA%D5%B7%D1%C5%E3%C1%C4%BA%C5%C2%EB%A1%BE%2B%F6%A6%5F%C5%AE%71%3A%32%31%37%39%37%30%36%39%33%36%A1%BF",
        "jd",
    )


def test_parse_query_jd_sha_tan_ku_nan_1446537277():
    verify_serp_parsing(
        "https://web.archive.org/web/20151103075437id_/http://search.jd.com/search?keyword=%E6%B2%99%E6%BB%A9%E8%A3%A4%20%E7%94%B7&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&sttr=1&ev=exbrand_Maro%20Homme%40&uc=0",
        "jd",
    )


def test_parse_query_jd_ji_guang_da_yin_1516054404():
    verify_serp_parsing(
        "https://web.archive.org/web/20180115221324id_/http://search.jd.com/search?keyword=%E6%BF%80%E5%85%89%E6%89%93%E5%8D%B0&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&wq=%E6%BF%80%E5%85%89%E6%89%93%E5%8D%B0&ev=5345_88976%40268435461_268466585%40&uc=0",
        "jd",
    )


def test_parse_query_jd_dao_tian_1442736754():
    verify_serp_parsing(
        "https://web.archive.org/web/20150920081234id_/http://search.jd.com/search?keyword=%E7%A8%BB%E7%94%B0&enc=utf-8&qrst=1&et=5&rt=1&stop=1&vt=2&ev=exbrand_%E7%A8%BB%E7%94%B0%EF%BC%88DENTSTAR%EF%BC%89%40&uc=0",
        "jd",
    )


def test_parse_query_jd_song_yan_1434320372():
    verify_serp_parsing(
        "https://web.archive.org/web/20150614221932id_/http://search.jd.com/search?keyword=%E5%AE%8B%E5%BD%A6&cid=3267&book=y",
        "jd",
    )


def test_parse_query_jd_yxingaiujliu_xwcp198_1537926187():
    verify_serp_parsing(
        "https://web.archive.org/web/20180926014307id_/http://search.jd.com/search?keyword=%C5%B7%C3%C0xingai%D3%B0%D2%F4%CF%C8%B7%E6%A1%BA%2B%CE%A2%D0%C5xwcp198%A1%BB%A1%D1&qr=&qrst=UNEXPAND&et=&rt=1&ev=exprice_0-19%40&uc=0",
        "jd",
    )


def test_parse_query_jd_mo_yan_1516163665():
    verify_serp_parsing(
        "https://web.archive.org/web/20180117043425id_/http://search.jd.com/search?keyword=%E8%8E%AB%E8%A8%80&enc=utf-8&qrst=1&rt=1&stop=1&book=y&vt=2&ev=publishers_%E4%B8%AD%E5%9B%BD%E5%8D%8E%E4%BE%A8%E5%87%BA%E7%89%88%E7%A4%BE%40package_%E5%B9%B3%E8%A3%85%40&uc=0",
        "jd",
    )


def test_parse_query_jd_you_xi_1511265243():
    verify_serp_parsing(
        "https://web.archive.org/web/20171121115403id_/http://search.jd.com/search?keyword=%E6%B8%B8%E6%88%8F&enc=utf-8&qrst=1&rt=1&stop=1&vt=2&offset=3&wq=%E6%B8%B8%E6%88%8F&cid2=3287&ev=packstate_no%402562_22800%40&uc=0",
        "jd",
    )



from archive_query_log.legacy.config import SERVICES_PATH
from archive_query_log.legacy.services import read_services


def test_services_can_be_parsed():
    assert read_services(
        SERVICES_PATH,
        ignore_parsing_errors=False
    ) is not None



"""
This script registers the Archive Query Log dataset in ir_datasets.
See: https://github.com/allenai/ir_datasets/
Note: This script is intended to be executed with the Docker image provided.
"""

from pathlib import Path
from typing import NamedTuple, List, Mapping, Optional, Iterator

from ir_datasets import registry
from ir_datasets.datasets.base import Dataset
from ir_datasets.formats import BaseDocs, BaseQueries, BaseQrels, GenericDoc
from ir_datasets.indices import PickleLz4FullStore, Docstore
from pandas import DataFrame, read_json, concat
from tqdm import tqdm

NAME = "archive-query-log"
DATA_DIR = Path('/data/')


def _extract_non_empty_results(serp: Mapping) -> List[dict]:
    if 'serp_results' not in serp or serp['serp_results'] is None:
        return []
    results = []
    for result in serp['serp_results']:
        if ('result_snippet_title' not in result or
                result['result_snippet_title'] is None):
            continue
        results += [result]
    return results


def _parse_serps(base_path: Path, min_results_on_serp: int = 1) -> DataFrame:
    result: DataFrame = concat([
        read_json(path, lines=True)
        for path in tqdm(base_path.glob('serps/part*.gz'), 'Load SERPs')
    ])
    original_length = len(result)
    num_results = result['serp_results'].map(
        lambda serp_results: len(
            _extract_non_empty_results({'serp_results': serp_results})
        )
    )
    mask = num_results >= min_results_on_serp
    result = result[mask]

    print(f'Processed {original_length} SERPs, '
          f'found {len(result)} non-empty SERPs.')

    return result


def _persist_run_file(serps: DataFrame, lang: str = "en") -> None:
    run_data = []
    for _, serp in serps.iterrows():
        results = _extract_non_empty_results(serp)
        max_rank = max(
            result['result_snippet_rank']
            for result in results
        )
        run_data += [
            {
                'query': serp.serp_id,
                'q0': 0,
                'docid': result['result_id'],
                'rank': result['result_snippet_rank'],
                'score': max_rank - result['result_snippet_rank'],
                'system': 'aql'
            }
            for result in results
            if serp['serp_query_text_url_language'] == lang
        ]

    run = DataFrame(run_data)
    run.to_csv(DATA_DIR / 'run.txt', sep=" ", header=False, index=False)


class ArchiveQueryLogQuery(NamedTuple):
    query_id: str
    text: str
    search_provider_name: str
    num_results: int
    lang: str

    def default_text(self):
        return self.text


class ArchiveQueryLogQrel(NamedTuple):
    query_id: str
    doc_id: str
    relevance: int
    iteration: int
    search_provider_name: str
    num_results: int
    lang: str


class ArchiveQueryLogQueries(BaseQueries):
    serps: DataFrame
    lang: Optional[str]

    def __init__(self, serps: DataFrame, lang: Optional[str] = "en"):
        self.serps = serps
        self.lang = lang

    def queries_iter(self) -> Iterator[ArchiveQueryLogQuery]:
        for _, serp in self.serps.iterrows():
            if (self.lang is not None and
                    self.lang != serp["serp_query_text_url_language"]):
                continue
            num_results = len(_extract_non_empty_results(serp))
            if num_results <= 0:
                continue
            yield ArchiveQueryLogQuery(
                query_id=serp["serp_id"],
                text=serp["serp_query_text_html"],
                lang=serp["serp_query_text_url_language"],
                search_provider_name=serp["search_provider_name"],
                num_results=num_results,
            )

    def queries_cls(self):
        return ArchiveQueryLogQuery

    def queries_lang(self) -> Optional[str]:
        return self.lang


class ArchiveQueryLogQrels(BaseQrels):
    serps: DataFrame
    lang: Optional[str]

    def __init__(self, serps: DataFrame, lang: Optional[str] = "en"):
        self.serps = serps
        self.lang = lang

    def qrels_iter(self) -> Iterator[ArchiveQueryLogQrel]:
        for _, serp in self.serps.iterrows():
            if (self.lang is not None and
                    self.lang != serp["serp_query_text_url_language"]):
                continue
            results = _extract_non_empty_results(serp)
            max_rank = max(
                result['result_snippet_rank']
                for result in results
            )
            for result in results:
                yield ArchiveQueryLogQrel(
                    query_id=serp["serp_id"],
                    doc_id=result['result_id'],
                    relevance=max_rank - result['result_snippet_rank'],
                    iteration=0,
                    search_provider_name=serp["search_provider_name"],
                    num_results=len(results),
                    lang=serp["serp_query_text_url_language"],
                )

    def qrels_cls(self):
        return ArchiveQueryLogQrel


class ArchiveQueryLogDocs(BaseDocs):
    serps: DataFrame
    lang: Optional[str]

    def __init__(self, serps: DataFrame, lang: Optional[str] = "en"):
        self.serps = serps
        self.lang = lang

    def docs_iter(self) -> Iterator[GenericDoc]:
        for _, serp in self.serps.iterrows():
            if (self.lang is not None and
                    self.lang != serp["serp_query_text_url_language"]):
                continue
            for result in _extract_non_empty_results(serp):
                yield GenericDoc(
                    doc_id=result['result_id'],
                    text=f"{result['result_snippet_title']} "
                         f"{result['result_snippet_text']}".strip()
                )

    def docs_count(self) -> int:
        return sum(
            len(_extract_non_empty_results(serp))
            for _, serp in self.serps.iterrows()
            if (self.lang is None or
                self.lang == serp["serp_query_text_url_language"])
        )

    def docs_cls(self):
        return GenericDoc

    def docs_lang(self) -> Optional[str]:
        return self.lang

    def docs_store(self) -> Docstore:
        return PickleLz4FullStore(
            path=f'{DATA_DIR}/docs.pklz4',
            init_iter_fn=self.docs_iter,
            data_cls=self.docs_cls(),
            lookup_field='doc_id',
            index_fields=['doc_id'],
        )


_serps = _parse_serps(DATA_DIR)

dataset = Dataset(
    ArchiveQueryLogDocs(_serps),
    ArchiveQueryLogQueries(_serps),
    ArchiveQueryLogQrels(_serps),
)

assert dataset.has_docs(), "dataset has no documents"
assert dataset.has_queries(), "dataset has no queries"

if NAME in registry:
    print(f"Dataset '{NAME}' already registered.")
else:
    registry.register(NAME, dataset)

if __name__ == '__main__':
    _persist_run_file(_serps)



from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack
from csv import writer
from datetime import datetime
from gzip import GzipFile
from pathlib import Path
from typing import Collection
from uuid import UUID

from click import option, BOOL, command
from tqdm.auto import tqdm

from archive_query_log.legacy import DATA_DIRECTORY_PATH
from archive_query_log.legacy.cli.util import PathParam
from archive_query_log.legacy.index import ArchivedRawSerpIndex, \
    ArchivedUrlIndex, ArchivedQueryUrlIndex, ArchivedParsedSerpIndex, \
    ArchivedSearchResultSnippetIndex, ArchivedRawSearchResultIndex, \
    LocatedRecord
from archive_query_log.legacy.model import ArchivedUrl, CorpusQueryUrl, \
    ArchivedSearchResultSnippet, CorpusDocument, CorpusJsonlLocation, \
    CorpusWarcLocation, ArchivedRawSerp, \
    ArchivedQueryUrl, ArchivedParsedSerp, CorpusQuery, CorpusSearchResult


@command(
    "corpus",
    help="Generate corpus.",
)
@option(
    "-d", "--data-directory", "--data-directory-path",
    type=PathParam(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=False,
        resolve_path=True,
        path_type=Path,
    ),
    default=DATA_DIRECTORY_PATH
)
@option(
    "-f", "--focused",
    type=BOOL,
    default=False,
    is_flag=True,
)
@option(
    "-q", "--queries",
    type=BOOL,
    default=False,
    is_flag=True,
)
@option(
    "-o", "--output-directory", "--output-directory-path",
    type=PathParam(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        path_type=Path,
    ),
    required=False,
)
def corpus_command(
        data_directory: Path,
        focused: bool,
        queries: bool,
        output_directory: Path,
) -> None:
    output_path: Path
    if output_directory is not None:
        output_path = output_directory
    elif focused:
        output_path = data_directory / "focused" / "corpus"
    else:
        output_path = data_directory / "corpus"
    output_path.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    queries_path = output_path / f"queries-{timestamp}.jsonl.gz"
    queries_path.touch(exist_ok=True)
    queries_offsets_path = output_path / f"queries-{timestamp}.jsonl.offsets"
    queries_offsets_path.touch(exist_ok=True)
    documents_path = output_path / f"documents-{timestamp}.jsonl.gz"
    documents_path.touch(exist_ok=True)
    documents_offsets_path = \
        output_path / f"documents-{timestamp}.jsonl.offsets"
    documents_offsets_path.touch(exist_ok=True)

    # Load indices.
    with ExitStack() as stack:
        archived_url_index = stack.enter_context(
            ArchivedUrlIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        archived_query_url_index = stack.enter_context(
            ArchivedQueryUrlIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        archived_raw_serp_index = stack.enter_context(
            ArchivedRawSerpIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        archived_parsed_serp_index = stack.enter_context(
            ArchivedParsedSerpIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        archived_search_result_snippet_index = stack.enter_context(
            ArchivedSearchResultSnippetIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        archived_raw_search_result_index = stack.enter_context(
            ArchivedRawSearchResultIndex(
                data_directory=data_directory,
                focused=focused,
            )
        )
        # archived_parsed_search_result_index = stack.enter_context(
        #     ArchivedParsedSearchResultIndex(
        #         data_directory=data_directory,
        #         focused=focused,
        #     )
        # )

        query_schema = CorpusQuery.schema()
        document_schema = CorpusDocument.schema()
        with queries_path.open("wb") as queries_file, \
                queries_offsets_path.open("w") as queries_offsets_file, \
                documents_path.open("wb") as documents_file, \
                documents_offsets_path.open("w") as documents_offsets_file:
            queries_offsets_writer = writer(queries_offsets_file)
            documents_offsets_writer = writer(documents_offsets_file)

            archived_ids: Collection[UUID]
            if queries:
                archived_ids = set(archived_query_url_index)
            else:
                archived_ids = set(archived_url_index)

            # noinspection PyTypeChecker
            archived_ids = tqdm(
                archived_ids,
                desc="Build corpus",
                unit="ID",
            )

            for archived_id in archived_ids:
                query_documents = _build_query_documents(
                    archived_url_index=archived_url_index,
                    archived_query_url_index=archived_query_url_index,
                    archived_raw_serp_index=archived_raw_serp_index,
                    archived_parsed_serp_index=archived_parsed_serp_index,
                    archived_search_result_snippet_index=(
                        archived_search_result_snippet_index
                    ),
                    archived_raw_search_result_index=(
                        archived_raw_search_result_index
                    ),
                    # archived_parsed_search_result_index=(
                    #     archived_parsed_search_result_index
                    # ),
                    archived_id=archived_id,
                )
                if query_documents is None:
                    continue
                query, documents = query_documents
                queries_file_offset = queries_file.tell()
                with GzipFile(
                        fileobj=queries_file,
                        mode="w",
                ) as queries_gzip_file:
                    queries_gzip_file.write(
                        f"{query_schema.dumps(query)}\n".encode("utf8")
                    )
                queries_offsets_writer.writerow(
                    [str(query.id), str(queries_file_offset)]
                )
                for document in documents:
                    documents_file_offset = documents_file.tell()
                    with GzipFile(
                            fileobj=documents_file,
                            mode="w",
                    ) as documents_gzip_file:
                        documents_gzip_file.write(
                            f"{document_schema.dumps(document)}\n".encode(
                                "utf8")
                        )
                    documents_offsets_writer.writerow(
                        [str(document.id), str(documents_file_offset)]
                    )


def _build_query_url(
        archived_url_loc: LocatedRecord[
            CorpusJsonlLocation, ArchivedUrl
        ],
        archived_query_url_loc: LocatedRecord[
                                    CorpusJsonlLocation, ArchivedQueryUrl
                                ] | None,
        archived_raw_serp_loc: LocatedRecord[
                                   CorpusWarcLocation, ArchivedRawSerp
                               ] | None,
        archived_parsed_serp_loc: LocatedRecord[
                                      CorpusJsonlLocation, ArchivedParsedSerp
                                  ] | None,
) -> CorpusQueryUrl:
    archived_url = archived_url_loc.record
    return CorpusQueryUrl(
        id=archived_url.id,
        url=archived_url.url,
        timestamp=archived_url.timestamp,
        wayback_url=archived_url.archive_url,
        wayback_raw_url=archived_url.raw_archive_url,
        url_query=(
            archived_query_url_loc.record.query
            if archived_query_url_loc is not None else None
        ),
        url_page=(
            archived_query_url_loc.record.page
            if archived_query_url_loc is not None else None
        ),
        url_offset=(
            archived_query_url_loc.record.offset
            if archived_query_url_loc is not None else None
        ),
        serp_query=(
            archived_parsed_serp_loc.record.interpreted_query
            if archived_parsed_serp_loc is not None else None
        ),
        archived_url_location=(
            archived_url_loc.location
            if archived_url_loc is not None else None
        ),
        archived_query_url_location=(
            archived_query_url_loc.location
            if archived_query_url_loc is not None else None
        ),
        archived_raw_serp_location=(
            archived_raw_serp_loc.location
            if archived_raw_serp_loc is not None else None
        ),
        archived_parsed_serp_location=(
            archived_parsed_serp_loc.location
            if archived_parsed_serp_loc is not None else None
        ),
    )


def _build_search_result(
        archived_search_result_snippet_index: ArchivedSearchResultSnippetIndex,
        archived_raw_search_result_index: ArchivedRawSearchResultIndex,
        # archived_parsed_search_result_index: ArchivedParsedSearchResultIndex,
        archived_search_result_snippet: ArchivedSearchResultSnippet,
) -> CorpusSearchResult:
    archived_snippet_loc = archived_search_result_snippet_index[
        archived_search_result_snippet.id]
    archived_raw_search_result_loc = archived_raw_search_result_index \
        .get(archived_search_result_snippet.id)
    # archived_parsed_search_result_loc = archived_parsed_search_result_index \
    #     .get(archived_search_result_snippet.id)
    # archived_parsed_search_result_loc = None
    return CorpusSearchResult(
        id=archived_search_result_snippet.id,
        url=archived_search_result_snippet.url,
        timestamp=archived_search_result_snippet.timestamp,
        wayback_url=archived_search_result_snippet.archive_url,
        wayback_raw_url=archived_search_result_snippet.raw_archive_url,
        snippet_rank=archived_search_result_snippet.rank,
        snippet_title=archived_search_result_snippet.title,
        snippet_text=archived_search_result_snippet.snippet,
        archived_snippet_location=archived_snippet_loc.location,
        archived_raw_search_result_location=(
            archived_raw_search_result_loc.location
            if archived_raw_search_result_loc is not None else None
        ),
        archived_parsed_search_result_location=None,
        # archived_parsed_search_result_location=(
        #     archived_parsed_search_result_loc.location
        #     if archived_parsed_search_result_loc is not None else None
        # ),
    )


def _build_query_documents(
        archived_url_index: ArchivedUrlIndex,
        archived_query_url_index: ArchivedQueryUrlIndex,
        archived_raw_serp_index: ArchivedRawSerpIndex,
        archived_parsed_serp_index: ArchivedParsedSerpIndex,
        archived_search_result_snippet_index: ArchivedSearchResultSnippetIndex,
        archived_raw_search_result_index: ArchivedRawSearchResultIndex,
        # archived_parsed_search_result_index: ArchivedParsedSearchResultIndex,
        archived_id: UUID,
) -> tuple[CorpusQuery, list[CorpusDocument]] | None:
    archived_url_loc = archived_url_index.get(archived_id)
    archived_query_url_loc = archived_query_url_index.get(archived_id)
    archived_raw_serp_loc = archived_raw_serp_index.get(archived_id)
    archived_parsed_serp_loc = archived_parsed_serp_index.get(archived_id)
    if archived_url_loc is None:
        return None
    query_url = _build_query_url(
        archived_url_loc,
        archived_query_url_loc,
        archived_raw_serp_loc,
        archived_parsed_serp_loc,
    )
    snippets = archived_parsed_serp_loc.record.results \
        if archived_parsed_serp_loc is not None else []
    path_components = archived_url_loc.location.relative_path.parts
    service = path_components[2] \
        if path_components[0] == "focused" else path_components[1]

    def convert_search_result(
            archived_search_result_snippet: ArchivedSearchResultSnippet
    ) -> CorpusSearchResult:
        return _build_search_result(
            archived_search_result_snippet_index,
            archived_raw_search_result_index,
            # archived_parsed_search_result_index,
            archived_search_result_snippet,
        )

    with ThreadPoolExecutor() as executor:
        results = list(executor.map(convert_search_result, snippets))
    query = CorpusQuery(
        service=service,
        id=query_url.id,
        url=query_url.url,
        timestamp=query_url.timestamp,
        wayback_url=query_url.wayback_url,
        wayback_raw_url=query_url.wayback_raw_url,
        url_query=query_url.url_query,
        url_page=query_url.url_page,
        url_offset=query_url.url_offset,
        serp_query=query_url.serp_query,
        archived_url_location=query_url.archived_url_location,
        archived_query_url_location=query_url.archived_query_url_location,
        archived_raw_serp_location=query_url.archived_raw_serp_location,
        archived_parsed_serp_location=query_url.archived_parsed_serp_location,
        results=results,
    )

    documents = [
        CorpusDocument(
            service=service,
            id=result.id,
            url=result.url,
            timestamp=result.timestamp,
            wayback_url=result.wayback_url,
            wayback_raw_url=result.wayback_raw_url,
            query=query,
            snippet_rank=result.snippet_rank,
            snippet_title=result.snippet_title,
            snippet_text=result.snippet_text,
            archived_snippet_location=result.archived_snippet_location,
            archived_raw_search_result_location=(
                result.archived_raw_search_result_location
            ),
            archived_parsed_search_result_location=None,
            # archived_parsed_search_result_location=(
            #     result.archived_parsed_search_result_location
            # ),
        )
        for result in results
    ]

    return query, documents



from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import cached_property
from hashlib import md5
from pathlib import Path
from typing import Sequence, Any
from urllib.parse import SplitResult, urlsplit
from uuid import UUID, uuid5, NAMESPACE_URL

from dataclasses_json import DataClassJsonMixin, config
from marshmallow.fields import List, Nested, String, Field
from publicsuffixlist import PublicSuffixList

from archive_query_log.legacy.model.highlight import HighlightedText
from archive_query_log.legacy.util.serialization import HighlightedTextField


@dataclass(frozen=True, slots=True)
class ArchivedUrl(DataClassJsonMixin):
    """
    A URL that is archived in the Wayback Machine (https://web.archive.org/).
    The archived snapshot can be retrieved using the ``archive_url``
    and ``raw_archive_url`` properties.

    Output of: 2-archived-urls
    Input of: 3-archived-query-urls
    """

    url: str
    """
    Original URL that was archived.
    """
    timestamp: int
    """
    Timestamp of the archived snapshot in the Wayback Machine.
    """

    @cached_property
    def id(self) -> UUID:
        """
        Unique ID for this archived URL.
        """
        return uuid5(NAMESPACE_URL, f"{self.timestamp}:{self.url}")

    @cached_property
    def split_url(self) -> SplitResult:
        """
        Original URL split into its components.
        """
        return urlsplit(self.url)

    @cached_property
    def url_domain(self) -> str:
        """
        Domain of the original URL.
        """
        return self.split_url.netloc

    @cached_property
    def url_md5(self) -> str:
        """
        MD5 hash of the original URL.
        """
        return md5(self.url.encode(), usedforsecurity=False).hexdigest()

    @cached_property
    def datetime(self) -> datetime:
        """
        Snapshot timestamp as a ``datetime`` object.
        """
        return datetime.fromtimestamp(self.timestamp, timezone.utc)

    @cached_property
    def archive_timestamp(self) -> str:
        """
        Snapshot timestamp as a string in the format used
        by the Wayback Machine (``YYYYmmddHHMMSS``).
        """
        return f"{self.datetime.year:04d}{self.datetime.month:02d}" \
               f"{self.datetime.day:02d}{self.datetime.hour:02d}" \
               f"{self.datetime.minute:02d}{self.datetime.second:02d}"

    @property
    def archive_url(self) -> str:
        """
        URL of the archived snapshot in the Wayback Machine.
        """
        return f"https://web.archive.org/web/" \
               f"{self.archive_timestamp}/{self.url}"

    @property
    def raw_archive_url(self) -> str:
        """
        URL of the archived snapshot's raw contents in the Wayback Machine.
        """
        return f"https://web.archive.org/web/" \
               f"{self.archive_timestamp}id_/{self.url}"


@dataclass(frozen=True, slots=True)
class ArchivedQueryUrl(ArchivedUrl, DataClassJsonMixin):
    """
    Archived URL of a search engine result page (SERP) for a query.

    Output of: 3-archived-query-urls
    Input of: 4-archived-raw-serps, 8-ir-corpus
    """

    query: str
    """
    Query that was used to retrieve the SERP.
    """
    page: int | None
    """
    Page number of the SERP, e.g., 1, 2, 3.

    Note: the page number should be zero-indexed, i.e.,
    the first result page has the page number 0.

    See also: ``results_page_offset``.
    """
    offset: int | None
    """
    Offset (start position) of the first result in the SERP, e.g., 10, 20.

    Note: the offset should be zero-indexed, i.e.,
    the first result page has the offset 0.

    See also ``results_page``.
    """


@dataclass(frozen=True, slots=True)
class ArchivedRawSerp(ArchivedQueryUrl, DataClassJsonMixin):
    """
    Raw snapshot content of an archived SERP.

    Output of: 4-archived-raw-serps
    Input of: 5-archived-parsed-serps
    """

    content: bytes
    """
    Raw byte content of the archived SERP's snapshot.
    """
    encoding: str
    """
    Encoding of the archived SERP's snapshot.
    """


@dataclass(frozen=True, slots=True)
class ArchivedSearchResultSnippet(ArchivedUrl, DataClassJsonMixin):
    """
    Single retrieved result from a query's archived SERP.
    """

    rank: int
    """
    Rank of the result in the SERP.
    """
    title: HighlightedText | str = field(metadata=config(
        encoder=str,
        decoder=HighlightedText,
        mm_field=HighlightedTextField(),
    ))
    """
    Title of the result with optional highlighting.
    """
    snippet: HighlightedText | str | None = field(metadata=config(
        encoder=str,
        decoder=HighlightedText,
        mm_field=HighlightedTextField(allow_none=True),
    ))
    """
    Snippet of the result.
    Highlighting should be normalized to ``<em>`` tags. 
    Other HTML tags are removed.
    """

    @cached_property
    def id(self) -> UUID:
        """
        Unique ID for this archived URL.
        """
        return uuid5(NAMESPACE_URL, f"{self.rank}:{self.timestamp}:{self.url}")


@dataclass(frozen=True, slots=True)
class ArchivedParsedSerp(ArchivedQueryUrl, DataClassJsonMixin):
    """
    Archived search engine result page (SERP) corresponding to a query.

    Output of: 5-archived-parsed-serps
    Input of: 6-archived-raw-search-results, 8-ir-corpus
    """

    results: Sequence[ArchivedSearchResultSnippet] = field(
        metadata=config(
            encoder=list,
            decoder=tuple,
            mm_field=List(Nested(ArchivedSearchResultSnippet.schema()))
        )
    )
    """
    Retrieved results from the SERP in the same order as they appear.
    """
    interpreted_query: str | None
    """
    Interpreted query that is displayed or otherwise included in the SERP.

    Note: the interpreted result query can be different from the original query
    due to spelling correction etc.
    """


@dataclass(frozen=True, slots=True)
class ArchivedRawSearchResult(ArchivedSearchResultSnippet, DataClassJsonMixin):
    """
    Raw content of an archived search result.

    Output of: 6-archived-raw-search-results
    Input of: 7-archived-parsed-search-results
    """

    content: bytes
    """
    Raw byte content of the archived search result's snapshot.
    """
    encoding: str
    """
    Encoding of the archived search result's snapshot.
    """


@dataclass(frozen=True, slots=True)
class ArchivedParsedSearchResult(ArchivedSearchResultSnippet,
                                 DataClassJsonMixin):
    """
    Parsed representation of an archived search result.

    Output of: 7-archived-parsed-search-results
    Input of: 8-ir-corpus
    """
    content_title: str | None
    """
    Title of the archived SERP's snapshot content.

    Note: the content title can be different from the snippet title due to ellipses etc.
    """
    content_plaintext: str | None
    """
    Plaintext of the archived SERP's snapshot content.
    """


# flake8: noqa: E402
from archive_query_log.legacy.model.parse import QueryParser, \
    PageParser, OffsetParser, QueryParserField, PageOffsetParserField, \
    ResultsParserField, InterpretedQueryParserField, InterpretedQueryParser, \
    ResultsParser

_public_suffix_list = PublicSuffixList()


@dataclass(frozen=True, slots=True)
class Service(DataClassJsonMixin):
    """
    An online service that has a search interface.

    Output of: service collection, service domains
    Input of: service URLs, query extraction
    """

    name: str
    """
    Service name (corresponds to ``alexa_domain`` without
    the ``alexa_public_suffix``).
    """
    domains: Sequence[str] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(String())
        ),
    )
    """
    Known domains of the service, including the main domain.
    """
    focused_url_prefixes: Sequence[str] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(String())
        ),
        default=(),
    )
    """
    URL prefixes for a more focused pipeline which might miss some queries
    but executes faster.
    """
    query_parsers: Sequence[QueryParser] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(QueryParserField())
        ),
        default=(),
    )
    """
    Query parsers in order of precedence.
    """
    page_parsers: Sequence[PageParser] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(PageOffsetParserField())
        ),
        default=(),
    )
    """
    Page number parsers in order of precedence.
    """
    offset_parsers: Sequence[OffsetParser] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(PageOffsetParserField())
        ),
        default=(),
    )
    """
    Page number parsers in order of precedence.
    """
    interpreted_query_parsers: Sequence[InterpretedQueryParser] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(InterpretedQueryParserField())
        ),
        default=(),
    )
    """
    Interpreted query parsers in order of precedence.
    The interpreted query is the query that is displayed 
    or otherwise included in the SERP.

    Note: the interpreted result query can be different from the original query
    due to spelling correction etc.
    """
    results_parsers: Sequence[ResultsParser] = field(
        metadata=config(
            decoder=tuple,
            mm_field=List(ResultsParserField())
        ),
        default=(),
    )
    """
    SERP parsers in order of precedence.
    """


class PathField(Field):
    def _serialize(
            self, value: Any, attr: str | None, obj: Any, **kwargs
    ) -> str:
        return str(value)

    def _deserialize(
            self, value: str, attr: str | None, data: Any, **kwargs: Any
    ) -> Path:
        return Path(value)


@dataclass(frozen=True, slots=True)
class CorpusJsonlLocation(DataClassJsonMixin):
    relative_path: Path = field(
        metadata=config(
            encoder=str,
            decoder=Path,
            mm_field=PathField(),
        )
    )
    """
    Path of the JSONL file relative to the corpus root path.
    """
    byte_offset: int
    """
    Position of the JSONL line's first byte in the decompressed JSONL file.
    """


@dataclass(frozen=True, slots=True)
class CorpusJsonlSnippetLocation(CorpusJsonlLocation, DataClassJsonMixin):
    index: int
    """
    Index of the snippet in the JSONL file entry's results list.
    """


@dataclass(frozen=True, slots=True)
class CorpusWarcLocation(DataClassJsonMixin):
    relative_path: Path = field(
        metadata=config(
            encoder=str,
            decoder=Path,
            mm_field=PathField(),
        )
    )
    """
    Path of the WARC file relative to the corpus root path.
    """
    byte_offset: int
    """
    Position of the WARC record's first byte in the compressed WARC file.
    """


@dataclass(frozen=True, slots=True)
class CorpusQueryUrl(DataClassJsonMixin):
    id: UUID
    """
    Unique ID based on the URL and timestamp.
    """
    url: str
    """
    Original URL that was archived.
    """
    timestamp: int
    """
    POSIX timestamp of the URL's archived snapshot in the Wayback Machine.
    """
    wayback_url: str
    """
    URL to access the archived snapshot in the Wayback Machine.
    """
    wayback_raw_url: str
    """
    URL to access the archived snapshot's raw contents in the Wayback Machine.
    """
    url_query: str | None
    """
    Query that was parsed from the URL.
    """
    url_page: int | None
    """
    SERP page number that was parsed from the URL, e.g., 1, 2, 3.

    Note: the page number should be zero-indexed, i.e.,
    the first result page has the page number 0.
    """
    url_offset: int | None
    """
    SERP results offset (start position) that was parsed from the URL, 
    e.g., 10, 20.

    Note: the offset should be zero-indexed, i.e.,
    the first result page has the offset 0.
    """
    serp_query: str | None
    """
    Interpreted query as displayed on (or otherwise included in) the SERP.

    Note: the interpreted result query can be different from the original query
    due to spelling correction etc.
    """
    archived_url_location: CorpusJsonlLocation
    """
    Location of the corresponding URL entry and JSONL file.
    """
    archived_query_url_location: CorpusJsonlLocation | None
    """
    Location of the corresponding query URL entry and JSONL file.
    """
    archived_raw_serp_location: CorpusWarcLocation | None
    """
    Location of the corresponding raw SERP entry and WARC file.
    """
    archived_parsed_serp_location: CorpusJsonlLocation | None
    """
    Location of the corresponding parsed SERP entry in the JSONL file.
    """


@dataclass(frozen=True, slots=True)
class CorpusSearchResult(DataClassJsonMixin):
    id: UUID
    """
    Unique ID for this archived URL.
    """
    url: str
    """
    Original URL that was archived.
    """
    timestamp: int
    """
    POSIX timestamp of the archived snapshot in the Wayback Machine.
    
    Note that there might not be a snapshot for the exact timestamp, 
    but the Wayback Machine will instead redirect 
    to the nearest available snapshot.
    """
    wayback_url: str
    """
    URL of the archived snapshot in the Wayback Machine.
    
    Note that there might not be a snapshot for the exact timestamp, 
    but the Wayback Machine will instead redirect 
    to the nearest available snapshot.
    """
    wayback_raw_url: str
    """
    URL of the archived snapshot's raw contents in the Wayback Machine.
    
    Note that there might not be a snapshot for the exact timestamp, 
    but the Wayback Machine will instead redirect 
    to the nearest available snapshot.
    """
    snippet_rank: int
    """
    Rank of the result in the SERP.
    """
    snippet_title: HighlightedText | str = field(metadata=config(
        encoder=str,
        decoder=HighlightedText,
        mm_field=HighlightedTextField(),
    ))
    """
    Snippet title of the result with optional highlighting.
    Highlighting should be normalized to ``<em>`` tags. 
    Other HTML tags are removed.
    """
    snippet_text: HighlightedText | str | None = field(metadata=config(
        encoder=str,
        decoder=HighlightedText,
        mm_field=HighlightedTextField(allow_none=True),
    ))
    """
    Snippet text of the result with optional highlighting.
    Highlighting should be normalized to ``<em>`` tags. 
    Other HTML tags are removed.
    """
    archived_snippet_location: CorpusJsonlSnippetLocation
    """
    Location of the corresponding snippet entry in the JSONL file.
    """
    archived_raw_search_result_location: CorpusWarcLocation | None
    """
    Location of the corresponding raw search result entry and WARC file.
    """
    archived_parsed_search_result_location: CorpusJsonlLocation | None
    """
    Location of the corresponding parsed search result entry in the JSONL file.
    """


@dataclass(frozen=True, slots=True)
class CorpusQuery(CorpusQueryUrl, DataClassJsonMixin):
    service: str
    """
    Name of the search engine service from which the query was fetched.
    """
    results: Sequence[CorpusSearchResult] | None = field(
        metadata=config(
            encoder=list,
            decoder=tuple,
            mm_field=List(Nested(CorpusSearchResult.schema()))
        )
    )
    """
    Retrieved results from the SERP in the same order as they appear.
    """


@dataclass(frozen=True, slots=True)
class CorpusDocument(CorpusSearchResult, DataClassJsonMixin):
    service: str
    """
    Name of the search engine service from which the snippet was fetched.
    """
    query: CorpusQueryUrl
    """
    Query and SERP that was used to retrieve this result.
    """



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_twitch_juliversal_1629750531():
    verify_serp_parsing(
        "https://web.archive.org/web/20210823202851id_/https://www.twitch.tv/search?term=juliversal",
        "twitch",
    )


def test_parse_query_twitch_simplevar_1662753027():
    verify_serp_parsing(
        "https://web.archive.org/web/20220909195027id_/https://www.twitch.tv/search?term=%09%20SimpleVar",
        "twitch",
    )


def test_parse_query_twitch_cpentagon_1640060311():
    verify_serp_parsing(
        "https://web.archive.org/web/20211221041831id_/https://www.twitch.tv/search?term=CPentagon",
        "twitch",
    )


def test_parse_query_twitch_rubberboy2001_1661349876():
    verify_serp_parsing(
        "https://web.archive.org/web/20220824140436id_/https://www.twitch.tv/search?term=rubberboy2001",
        "twitch",
    )


def test_parse_query_twitch_central_fluminense_1630540704():
    verify_serp_parsing(
        "https://web.archive.org/web/20210901235824id_/https://www.twitch.tv/search?term=central%20fluminense",
        "twitch",
    )


def test_parse_query_twitch_xxlillythefallenangelneko_1638019769():
    verify_serp_parsing(
        "https://web.archive.org/web/20211127132929id_/https://www.twitch.tv/search?term=xxlillythefallenangelneko",
        "twitch",
    )


def test_parse_query_twitch_a_plague_tale_innocence_1638358324():
    verify_serp_parsing(
        "https://web.archive.org/web/20211201113204id_/https://www.twitch.tv/search?term=A%20Plague%20Tale%3A%20Innocence&type=categories&ref=NEXARDA",
        "twitch",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_twitter_rabble_ca_lang_ar_1442999332():
    verify_serp_parsing(
        "https://web.archive.org/web/20150923090852id_/https://mobile.twitter.com/search?vertical=default&near=me&q=rabble.ca%3Flang%3Dar&locale=ro",
        "twitter",
    )


def test_parse_query_twitter_rabble_ca_lang_fr_lang_id_1443046297():
    verify_serp_parsing(
        "https://web.archive.org/web/20150923221137id_/https://twitter.com/search?q=rabble.ca%3Flang%3Dfr%3Flang%3Did&f=videos&vertical=default&lang=sv",
        "twitter",
    )


def test_parse_query_twitter_lauramajor_1652839406():
    verify_serp_parsing(
        "https://web.archive.org/web/20220518020326id_/https://twitter.com/search?q=lauramajor&lang=no",
        "twitter",
    )


def test_parse_query_twitter_ecotourism_1557315321():
    verify_serp_parsing(
        "https://web.archive.org/web/20190508113521id_/https://twitter.com/search?q=%23Ecotourism",
        "twitter",
    )


def test_parse_query_twitter_freemariabutina_1563062689():
    verify_serp_parsing(
        "https://web.archive.org/web/20190714000449id_/https://twitter.com/search?q=%23FreeMariaButina&l=is&vertical=default&lang=mr",
        "twitter",
    )


def test_parse_query_twitter_http_shop_pre_com_corona02_p_409033_1648579256():
    verify_serp_parsing(
        "https://web.archive.org/web/20220329184056id_/https://mobile.twitter.com/search?q=http%3A%2F%2Fshop-pre.com%2Fcorona02%2F%3Fp%3D409033&lang=kn",
        "twitter",
    )


def test_parse_query_twitter_clubs_lang_sr_1591413117():
    verify_serp_parsing(
        "https://web.archive.org/web/20200606031157id_/https://twitter.com/search?q=clubs%3Flang%3Dsr&lang=en",
        "twitter",
    )


def test_parse_query_twitter_freemariabutina_1555124003():
    verify_serp_parsing(
        "https://web.archive.org/web/20190413025323id_/https://mobile.twitter.com/search?vertical=default&locale=es&q=%23FreeMariaButina&l=ja",
        "twitter",
    )


def test_parse_query_twitter_freemariabutina_1562498575():
    verify_serp_parsing(
        "https://web.archive.org/web/20190707112255id_/https://mobile.twitter.com/search?vertical=default&locale=id&q=%23FreeMariaButina&f=tweets",
        "twitter",
    )


def test_parse_query_twitter_campaigns_1481768285():
    verify_serp_parsing(
        "https://web.archive.org/web/20161215021805id_/https://twitter.com/search?q=campaigns&lang=ko",
        "twitter",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_pornhub_kardashian_1232246712():
    verify_serp_parsing(
        "https://web.archive.org/web/20090118024512id_/http://www.pornhub.com:80/video/search?search=kardashian&p=&c=32",
        "pornhub",
    )


def test_parse_query_pornhub_org_1219567054():
    verify_serp_parsing(
        "https://web.archive.org/web/20080824083734id_/http://www.pornhub.com:80/video/search?search=org",
        "pornhub",
    )


def test_parse_query_pornhub_mini_skirt_1257166127():
    verify_serp_parsing(
        "https://web.archive.org/web/20091102124847id_/http://www.pornhub.com:80/video/search?search=mini+skirt&c=37",
        "pornhub",
    )


def test_parse_query_pornhub_full_porno_films_1633148146():
    verify_serp_parsing(
        "https://web.archive.org/web/20211002041546id_/https://www.pornhub.com/video/search?search=full+porno+films",
        "pornhub",
    )


def test_parse_query_pornhub_brazilian_carnival_1378864891():
    verify_serp_parsing(
        "https://web.archive.org/web/20130911020131id_/http://www.pornhub.com:80/video/search?search=brazilian+carnival",
        "pornhub",
    )


def test_parse_query_pornhub_cfnm2_1271707129():
    verify_serp_parsing(
        "https://web.archive.org/web/20100419195849id_/http://www.pornhub.com:80/video/search?search=CFNM2&x=0&y=0",
        "pornhub",
    )


def test_parse_query_pornhub_12_1565436858():
    verify_serp_parsing(
        "https://web.archive.org/web/20190810113418id_/https://www.pornhub.com/video/search?search=12&o=mv",
        "pornhub",
    )


def test_parse_query_pornhub_roxy_reynolds_1222496976():
    verify_serp_parsing(
        "https://web.archive.org/web/20080927062936id_/http://www.pornhub.com:80/video/search?search=Roxy+Reynolds&c=1",
        "pornhub",
    )


def test_parse_query_pornhub_teen_facial_1237526884():
    verify_serp_parsing(
        "https://web.archive.org/web/20090320052804id_/http://www.pornhub.com:80/video/search?search=teen+facial&p=&c=15",
        "pornhub",
    )


def test_parse_query_pornhub_stripping_1627962626():
    verify_serp_parsing(
        "https://web.archive.org/web/20210803035026id_/https://www.pornhub.com/video/search?search=stripping%5D",
        "pornhub",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_google_taikoo_hui_mandarin_oriental_hotel_guangzhou_1652086766():
    verify_serp_parsing(
        "https://web.archive.org/web/20220509085926id_/https://www.google.com/search?q=Taikoo+Hui+Mandarin+Oriental+Hotel+Guangzhou",
        "google",
    )


def test_parse_query_google_finance_1540158323():
    verify_serp_parsing(
        "https://web.archive.org/web/20181021214523id_/http://www.google.co.uk/search?q=finance&gbv=1&sei=jw_NW8voLJWv0PEPld2DuA8",
        "google",
    )


def test_parse_query_google_hawaii_part_ii_lyrics_1633318830():
    verify_serp_parsing(
        "https://web.archive.org/web/20211004034030id_/https://www.google.com/search?q=Hawaii+part+ii+lyrics&sa=X&ved=2ahUKEwjA74_v6arzAhVTwZ4KHfAcBCAQ1QJ6BAgFEAw",
        "google",
    )


def test_parse_query_google_attack_on_titan_season_3_part_2_1556408967():
    verify_serp_parsing(
        "https://web.archive.org/web/20190427234927id_/https://www.google.com/search?q=attack+on+titan+season+3+part+2&rlz=1C1CHBF_en-GBGB760GB760&oq=attac&aqs=chrome.0.69i59j0l2j69i65l3.1486j0j4&sourceid=chrome&ie=UTF-8",
        "google",
    )


def test_parse_query_google_win10_iso_xia_zai_1577717811():
    verify_serp_parsing(
        "https://web.archive.org/web/20191230145651id_/https://www.google.com/search?q=win10+iso+%E4%B8%8B%E8%BD%BD&tbm=nws&source=lnt&tbs=qdr%3Ay&sa=X&ved=0ahUKEwjurrCb3t3mAhXPjp4KHT3MDxg4FBCnBQgf",
        "google",
    )


def test_parse_query_google_243_1628713922():
    verify_serp_parsing(
        "https://web.archive.org/web/20210811203202id_/https://www.google.com/search?q=243",
        "google",
    )


def test_parse_query_google_does_steve_has_a_beard_1601705030():
    verify_serp_parsing(
        "https://web.archive.org/web/20201003060350id_/https://www.google.com/search?q=does+steve+has+a+beard&oq=does+steve+has+a+beard&aqs=chrome..69i57j0i22i30l3.6767j0j4&sourceid=chrome&ie=UTF-8",
        "google",
    )


def test_parse_query_google_why_is_one_foot_slightly_larger_than_the_other_one_1605140430():
    verify_serp_parsing(
        "https://web.archive.org/web/20201112002030id_/https://www.google.com/search?q=Why%20is%20one%20foot%20slightly%20larger%20than%20the%20other%20one%3F",
        "google",
    )


def test_parse_query_google_zombie_apocalypse_1565114134():
    verify_serp_parsing(
        "https://web.archive.org/web/20190806175534id_/https://www.google.com/search?q=zombie+apocalypse&rlz=1C1GCEA_enGB797GB797&source=lnms&tbm=vid&sa=X&ved=0ahUKEwjo4aahg-_jAhUPTcAKHa1NCZ4Q_AUIEigC",
        "google",
    )


def test_parse_query_google_scholar_w_fan_j_li_s_ma_n_tang_and_w_yu_april_2012_towards_certain_fixes_with_editing_rules_and_master_data_the_vldb_journal_21_2_213_238_297_10_1007_s00778_011_0253_7_1614165399():
    verify_serp_parsing(
        "https://web.archive.org/web/20210224111639id_/https://scholar.google.com/scholar?hl=en&q=W.+Fan%2C+J.+Li%2C+S.+Ma%2C+N.+Tang%2C+and+W.+Yu.+April+2012.+Towards+certain+fixes+with+editing+rules+and+master+data.+The+VLDB+Journal%2C+21(2)%3A+213--238.+297+10.1007%2Fs00778-011-0253-7+",
        "google-scholar",
    )


def test_parse_query_google_susan_boyle_make_me_a_channel_of_your_peace_1607791072():
    verify_serp_parsing(
        "https://web.archive.org/web/20201212163752id_/http://www.google.ca/search?sa=X&q=Susan+Boyle+Make+Me+a+Channel+of+Your+Peace&stick=H4sIAAAAAAAAAIVUPWjUYBjOnbZe0xZ6qaDcIPF0qBW9Sy4_PZcWi0KhpdLe0ikkX_Ill3z58vOld0lEBOdOgjoKTg4dnRRECg4WcXAQXKy4OSm6dNHaWvOzmSHkyfu8z_t87_vy1UaaEy2zxZG2SAIIZX7cN1RgsMTDJtmtZLFIGRD_HwxEW9NEIhcQKAArOVmS-MDVdyuTR5ADHXdIBibIsJMocafLOxmGURpwsaJkWI_dQISWmWFbsNwUDt2imk3ECGZhLbHjMLLNwipydDvM1QU_4FwgFGHX9rtcFragrQ1DG5SyzVTUSmxFCkvarm5CkpceSG2lnfo5ngNSYA7S0sn9FEVR7oWXdEBwuzhJalkcKtQHWBasPJvHlmaRoi8waQucmxbZsQbNuCyGO7hoC4qDLhDzsBXITsqVBipBvhjZ8ewLGMaWGJR8CxiLYo6tyA263FyGTc4M7O4QlfghAFJpe0iSup2ScRzBbj5fTg-QFJS7JGNRlsrLFiOrEBuKtpLPACqQxwF2CrbAoaPtyNk8GQqlJoFYz5dDk2V-btDh85bLStJ1dLdkXI9BqXTqwJB8qbyujk993f803XhRvb_95kPlWZWeWvY8YqBkzUBqZOg9j7lIj97AUT9KmAY90ai13FZbEm1_hmrQdI6YrQo9vm5EPW_F0_swYe4wt-mxFcPVjJCsQqZH04seQgaI-h5mbtLNxrkWyH-0QgN4od7HpgLQJokOcw7lWfo_nObptb8GhBDB4y9ZgNG16ky16fDK8_cP341u1AdvDw4m_aWFxqVZhj61RJY9oKL6o89XXn38_W1-dpqu9dTYw56b1Km_z_f55vmxCz-Sx1vbe_P1n7MUtdc_s3D2BEvNUNzq_q-7Ow-uLsxS914-3Rk9WatMURNPRi6vbxIVs9e9BBnsiuocvgxWZRctFWMDsR5kN7zNkL11dCP9AYhSBnypBAAA&ved=2ahUKEwixvarn35TtAhXC3J4KHZIqBmYQ-BZ6BAgSEEA",
        "google",
    )


def test_parse_query_google_scholar_muhammed_rashid_1656890873():
    verify_serp_parsing(
        "https://web.archive.org/web/20220703232753id_/https://scholar.google.com/scholar?q=Muhammed+Rashid",
        "google-scholar",
    )


def test_parse_query_google_scholar_a_tumeo_m_branca_l_camerini_a_dual_priority_realtime_multiprocessor_system_on_fpga_for_automotive_ap_1614181186():
    verify_serp_parsing(
        "https://web.archive.org/web/20210224153946id_/https://scholar.google.com/scholar?hl=en&q=A.+Tumeo%2C+M.+Branca%2C+L.+Camerini%2C+%22A+Dual-Priority+RealTime+Multiprocessor+System+on+FPGA+for+Automotive+Applications%22%2C+Design%2C+Automation+and+Test+in+Europe+(DATE)+2008%2C+pages+1039--1044.+10.1145%2F1403375.1403625+",
        "google-scholar",
    )


def test_parse_query_google_hola_games_1552338270():
    verify_serp_parsing(
        "https://web.archive.org/web/20190311210430id_/https://www.google.com/search?q=Hola+Games&stick=H4sIAAAAAAAAAONgFuLSz9U3KE7PKCk2UOLVT9c3NEw3M801Kykr1lLOTrbSTy5L10_Ozy0oLUktii_LTEnNT0_MTbUqBrLKEyuLHzHGcQu8_HFPWCp00pqT1xj9uYjRJaTBxeaaV5JZUikkx8UnheQGDQYpHi4kPs8iVmGQ1mKFnMzsVIWSjFSFovz8XACNNlkjuQAAAA&sa=X&ved=0ahUKEwi839n-i_vgAhUGMt8KHUS9AjgQ-BYIUA",
        "google",
    )


def test_parse_query_google_chi_zhen_nzhong_rong_keteru_1608917828():
    verify_serp_parsing(
        "https://web.archive.org/web/20201225173708id_/https://www.google.co.jp/search?q=%E6%AD%AF+%E7%9C%9F%E3%82%93%E4%B8%AD%E6%BA%B6%E3%81%91%E3%81%A6%E3%82%8B&ie=UTF-8&oe=UTF-8&hl=ja-jp&client=safari",
        "google",
    )


def test_parse_query_google_bill_easley_sophisticated_prelude_1626498182():
    verify_serp_parsing(
        "https://web.archive.org/web/20210717050302id_/https://www.google.com/search?q=bill+easley+sophisticated+prelude&stick=H4sIAAAAAAAAAONgFuLSz9U3MM4wNSspUeLVT9c3NMwyM8jJMCvI0xLLTrbSzy0tzkzWTywqySwusSrOz0svXsSqmJSZk6OQmlick1qpUJxfkAGUy0xOLElNUSgoSs0pTUkFAB_nXk9ZAAAA&sa=X&ved=2ahUKEwjr3cn27-PxAhUTHzQIHXLBBTAQri4wCnoECAoQVA",
        "google",
    )


def test_parse_query_google_atalanta_bergamasca_calcio_wikipedia_1643703998():
    verify_serp_parsing(
        "https://web.archive.org/web/20220201082638id_/https://www.google.com/search?q=%22Atalanta+Bergamasca+Calcio%22+-wikipedia&source=lnms&tbm=vid&sa=X&ved=0ahUKEwirydi73J_1AhVCl2oFHdPMDeUQ_AUICygE",
        "google",
    )


def test_parse_query_google_orlando_bedoya_site_wikipedia_org_site_wikimedia_org_1629326434():
    verify_serp_parsing(
        "https://web.archive.org/web/20210818224034id_/https://www.google.com/search?safe=off&tbs=sur:fmc&tbm=isch&q=%22Orlando+bedoya%22+-site:wikipedia.org+-site:wikimedia.org",
        "google",
    )


def test_parse_query_google_flop_thumbs_up_1568839345():
    verify_serp_parsing(
        "https://web.archive.org/web/20190918204225id_/https://www.google.com/search?q=flop+thumbs+up&oq=flop+thumbs+up&aqs=chrome..69i57j69i60l2j69i61j69i65l2.3098j0j7&sourceid=chrome&ie=UTF-8",
        "google",
    )


def test_parse_query_google_ocolc_826746_1240020639():
    verify_serp_parsing(
        "https://web.archive.org/web/20090418021039id_/http://www.google.com/search?q=(OCoLC)826746&safe=vss&vss=1&sa=Google+Search",
        "google",
    )


def test_parse_query_google_george_w_summers_1637786317():
    verify_serp_parsing(
        "https://web.archive.org/web/20211124203837id_/https://www.google.com/search?as_eq=wikipedia&q=%22George+W.+Summers%22",
        "google",
    )


def test_parse_query_google_grasshopper_design_pvt_ltd_1640249780():
    verify_serp_parsing(
        "https://web.archive.org/web/20211223085620id_/https://www.google.com/search?as_eq=wikipedia&q=%22Grasshopper+Design+Pvt.+Ltd.%22",
        "google",
    )


def test_parse_query_google_pareques_site_hu_wikipedia_org_1647941362():
    verify_serp_parsing(
        "https://web.archive.org/web/20220322092922id_/https://www.google.com/search?q=%22Pareques%22+site:hu.wikipedia.org",
        "google",
    )


def test_parse_query_google_coneybury_anomaly_wikipedia_1640176109():
    verify_serp_parsing(
        "https://web.archive.org/web/20211222122829id_/https://www.google.com/search?tbs=bks:1&q=%22Coneybury+Anomaly%22+-wikipedia",
        "google",
    )


def test_parse_query_google_laxmikant_pyarelal_1657295550():
    verify_serp_parsing(
        "https://web.archive.org/web/20220708155230id_/https://www.google.com/search?as_eq=wikipedia&q=%22Laxmikant%E2%80%93Pyarelal%22",
        "google",
    )


def test_parse_query_google_kurabayashi_1643815267():
    verify_serp_parsing(
        "https://web.archive.org/web/20220202152107id_/https://www.google.com/search?as_eq=wikipedia&q=%22Kurabayashi%22",
        "google",
    )


def test_parse_query_google_gnosticism_1615175612():
    verify_serp_parsing(
        "https://web.archive.org/web/20210308035332id_/https://www.google.com/search?as_eq=wikipedia&q=%22Gnosticism%22",
        "google",
    )


def test_parse_query_google_fatmira_brecani_1662136538():
    verify_serp_parsing(
        "https://web.archive.org/web/20220902163538id_/https://www.google.com/search?as_eq=wikipedia&q=%22Fatmira+Bre%C3%A7ani%22",
        "google",
    )


def test_parse_query_google_kostiantyn_magaletskyi_1638351261():
    verify_serp_parsing(
        "https://web.archive.org/web/20211201093421id_/https://www.google.com/search?as_eq=wikipedia&q=%22Kostiantyn+Magaletskyi%22",
        "google",
    )


def test_parse_query_google_future_planning_1648480487():
    verify_serp_parsing(
        "https://web.archive.org/web/20220328151447id_/https://www.google.com/search?as_eq=wikipedia&q=%22Future+Planning%22",
        "google",
    )


def test_parse_query_google_list_of_surviving_lockheed_f_104_starfighters_1662745731():
    verify_serp_parsing(
        "https://web.archive.org/web/20220909174851id_/https://www.google.com/search?as_eq=wikipedia&q=%22List+of+surviving+Lockheed+F-104+Starfighters%22",
        "google",
    )


def test_parse_query_google_modern_times_chaplin_watch_online_1662343522():
    verify_serp_parsing(
        "https://web.archive.org/web/20220905020522id_/https://www.google.com.hk/search?q=modern+times+chaplin+watch+online",
        "google",
    )


def test_parse_query_google_lars_brygmann_1633135602():
    verify_serp_parsing(
        "https://web.archive.org/web/20211002004642id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lars+Brygmann%22",
        "google",
    )


def test_parse_query_google_boulevard_records_canada_site_wikipedia_org_site_wikimedia_org_1510135498():
    verify_serp_parsing(
        "https://web.archive.org/web/20171108100458id_/https://www.google.com/search?safe=off&tbm=isch&tbs=sur:fmc&q=%22Boulevard+Records%22+Canada+-site:wikipedia.org+-site:wikimedia.org",
        "google",
    )


def test_parse_query_google_continental_automotive_systems_wikipedia_1651015350():
    verify_serp_parsing(
        "https://web.archive.org/web/20220426232230id_/https://www.google.com/search?tbs=bks:1&q=%22Continental+Automotive+Systems%22+-wikipedia",
        "google",
    )


def test_parse_query_google_bouncing_hero_switch_site_easyallies_com_1633303267():
    verify_serp_parsing(
        "https://web.archive.org/web/20211003232107id_/https://www.google.com/search?q=%22Bouncing%20Hero%22+%22switch%22+site:easyallies.com",
        "google",
    )


def test_parse_query_google_evan_colborne_1647984326():
    verify_serp_parsing(
        "https://web.archive.org/web/20220322212526id_/https://www.google.com/search?q=%22Evan+Colborne%22&rlz=1C1GCEB_enCA943CA943&oq=%22Evan+Colborne%22&aqs=chrome..69i57j0i30.3299j0j7&sourceid=chrome&ie=UTF-8",
        "google",
    )


def test_parse_query_google_lycee_francais_louis_pasteur_de_lagos_1638831028():
    verify_serp_parsing(
        "https://web.archive.org/web/20211206225028id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lyc%C3%A9e+Fran%C3%A7ais+Louis+Pasteur+de+Lagos%22",
        "google",
    )


def test_parse_query_google_compress_tube_wikipedia_1644670761():
    verify_serp_parsing(
        "https://web.archive.org/web/20220212125921id_/https://www.google.com/search?tbs=bks:1&q=%22Compress+Tube%22+-wikipedia",
        "google",
    )


def test_parse_query_google_lige_gardner_1637042221():
    verify_serp_parsing(
        "https://web.archive.org/web/20211116055701id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lige+Gardner%22",
        "google",
    )


def test_parse_query_google_lope_recio_loynaz_1658853508():
    verify_serp_parsing(
        "https://web.archive.org/web/20220726163828id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lope+Recio+Loynaz%22",
        "google",
    )


def test_parse_query_google_bozsa_bence_site_hu_wikipedia_org_1647787452():
    verify_serp_parsing(
        "https://web.archive.org/web/20220320144412id_/https://www.google.com/search?q=%22Bozsa+Bence%22+site:hu.wikipedia.org",
        "google",
    )


def test_parse_query_google_fer_palacio_1631069534():
    verify_serp_parsing(
        "https://web.archive.org/web/20210908025214id_/https://www.google.com/search?as_eq=wikipedia&q=%22Fer+Palacio%22",
        "google",
    )


def test_parse_query_google_green_capital_energy_site_wikipedia_org_site_wikimedia_org_1657756151():
    verify_serp_parsing(
        "https://web.archive.org/web/20220713234911id_/https://www.google.com/search?safe=off&tbs=sur:fmc&tbm=isch&q=%22Green+Capital+Energy%22+-site:wikipedia.org+-site:wikimedia.org&spell=1&sa=X&ved=2ahUKEwie4oTthrz1AhWTHTQIHbsuDQ8QBXoECAAQAQ",
        "google",
    )


def test_parse_query_google_facebuster_1650197554():
    verify_serp_parsing(
        "https://web.archive.org/web/20220417121234id_/https://www.google.com/search?&q=%22Facebuster%22&tbs=bkt:s&tbm=bks",
        "google",
    )


def test_parse_query_google_focusrite_1641240556():
    verify_serp_parsing(
        "https://web.archive.org/web/20220103200916id_/https://www.google.com/search?as_eq=wikipedia&q=%22Focusrite%22",
        "google",
    )


def test_parse_query_google_kolex_1645947964():
    verify_serp_parsing(
        "https://web.archive.org/web/20220227074604id_/https://www.google.com/search?as_eq=wikipedia&q=%22Kolex%22",
        "google",
    )


def test_parse_query_google_list_of_grand_dukes_of_russia_1645733808():
    verify_serp_parsing(
        "https://web.archive.org/web/20220224201648id_/https://www.google.com/search?as_eq=wikipedia&q=%22List+of+grand+dukes+of+Russia%22",
        "google",
    )


def test_parse_query_google_green_wave_106_5_fm_1637281540():
    verify_serp_parsing(
        "https://web.archive.org/web/20211119002540id_/http://www.google.com/search?tbo=p&tbm=bks&q=%22Green+Wave+106.5+FM%22",
        "google",
    )


def test_parse_query_google_front_polisario_khat_al_shahid_1652852892():
    verify_serp_parsing(
        "https://web.archive.org/web/20220518054812id_/https://www.google.com/search?as_eq=wikipedia&q=%22Front+Polisario+Khat+al-Shahid%22",
        "google",
    )


def test_parse_query_google_guntury_1650179158():
    verify_serp_parsing(
        "https://web.archive.org/web/20220417070558id_/https://www.google.com/search?as_eq=wikipedia&q=%22Guntury%22",
        "google",
    )


def test_parse_query_google_film_review_annual_1660769248():
    verify_serp_parsing(
        "https://web.archive.org/web/20220817204728id_/https://www.google.com/search?as_eq=wikipedia&q=%22Film+Review+(annual)%22",
        "google",
    )


def test_parse_query_google_lucas_ruzin_1658326494():
    verify_serp_parsing(
        "https://web.archive.org/web/20220720141454id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lucas+Ruzin%22",
        "google",
    )


def test_parse_query_google_kush_raval_indian_cricketer_site_wikipedia_org_site_wikimedia_org_1649726560():
    verify_serp_parsing(
        "https://web.archive.org/web/20220412012240id_/https://www.google.com/search?safe=off&tbs=sur:fmc&tbm=isch&q=%22Kush+Raval+Indian+Cricketer%22+-site:wikipedia.org+-site:wikimedia.org",
        "google",
    )


def test_parse_query_google_sarah_monfort_wikipedia_1647374906():
    verify_serp_parsing(
        "https://web.archive.org/web/20220315200826id_/https://www.google.com/search?tbs=bks:1&q=%22Sarah+Monfort%22+-wikipedia",
        "google",
    )


def test_parse_query_google_list_of_crossings_of_the_james_river_1662064377():
    verify_serp_parsing(
        "https://web.archive.org/web/20220901203257id_/https://www.google.com/search?as_eq=wikipedia&q=%22List+of+crossings+of+the+James+River%22",
        "google",
    )


def test_parse_query_google_kothuru_assembly_constituency_1631384737():
    verify_serp_parsing(
        "https://web.archive.org/web/20210911182537id_/https://www.google.com/search?as_eq=wikipedia&q=%22Kothuru%22+Assembly+constituency",
        "google",
    )


def test_parse_query_google_can_ju_ying_yu_1560296360():
    verify_serp_parsing(
        "https://web.archive.org/web/20190611233920id_/https://www.google.com/search?safe=active&q=%E9%A4%90%E5%85%B7%E8%8B%B1%E8%AF%AD&sa=X&ved=2ahUKEwjhxJLa5-LiAhWRCXwKHSL2ArUQ1QIoB3oECAoQCA",
        "google",
    )


def test_parse_query_google_leo_de_la_strehaia_1633884656():
    verify_serp_parsing(
        "https://web.archive.org/web/20211010165056id_/https://www.google.com/search?as_eq=wikipedia&q=%22Leo+de+la+Strehaia%22",
        "google",
    )


def test_parse_query_google_conflict_middle_east_wikipedia_1636797028():
    verify_serp_parsing(
        "https://web.archive.org/web/20211113095028id_/https://www.google.com/search?tbs=bks:1&q=%22Conflict:+Middle+East%22+-wikipedia",
        "google",
    )


def test_parse_query_google_diffusion_systems_for_evaluation_of_biomineralization_1618448715():
    verify_serp_parsing(
        "https://web.archive.org/web/20210415010515id_/http://www.google.com/search?hl=en&ie=UTF-8&lr=&num=100&prmd=ivns&q=%22Diffusion+Systems+for+Evaluation+of+Biomineralization&sa=X&safe=off&site=unclesam&source=lnms&tbm=isch&ved=0ahUKEwiN-62-svLZAhXEl-AKHReSCO4Q_AUIBQ",
        "google",
    )


def test_parse_query_google_list_of_cantons_of_france_1634951028():
    verify_serp_parsing(
        "https://web.archive.org/web/20211023010348id_/https://www.google.com/search?as_eq=wikipedia&q=%22List+of+cantons+of+France%22",
        "google",
    )


def test_parse_query_google_scholar_wan_y_menon_s_and_ramaprasad_a_2009_the_paradoxical_nature_of_electronic_decision_aids_on_comparison_1614187144():
    verify_serp_parsing(
        "https://web.archive.org/web/20210224171904id_/https://scholar.google.com/scholar?hl=en&q=Wan%2C+Y.%2C+Menon%2C+S.%2C+and+Ramaprasad%2C+A.+2009.+The+paradoxical+nature+of+electronic+decision+aids+on+comparison-shopping%3A+The+experiments+and+analysis.+J.+Theoret.+Appl.+Electron.+Commerce+Res.%2C+80--96.+",
        "google-scholar",
    )


def test_parse_query_google_crow_names_1608438374():
    verify_serp_parsing(
        "https://web.archive.org/web/20201220042614id_/https://www.google.com/search?q=crow+names&rlz=1C1GCEA_enCA828CA828&oq=crow+name&aqs=chrome.0.0i457j69i57j0l6.1630j0j7&sourceid=chrome&ie=UTF-8",
        "google",
    )


def test_parse_query_google_giulia_botti_1660836463():
    verify_serp_parsing(
        "https://web.archive.org/web/20220818152743id_/https://www.google.com/search?as_eq=wikipedia&q=%22Giulia+Botti%22",
        "google",
    )


def test_parse_query_google_luiz_carlos_footballer_born_1980_wikipedia_1637131910():
    verify_serp_parsing(
        "https://web.archive.org/web/20211117065150id_/https://www.google.com/search?q=%22Luiz+Carlos%22+footballer,+born+1980+-wikipedia&gbv=1&sei=Gv2OYYXOMIHB-wTkppewDg",
        "google",
    )


def test_parse_query_google_now_it_s_time_for_your_treatment_relax_make_yourself_comfortable_1427746605():
    verify_serp_parsing(
        "https://web.archive.org/web/20150330201645id_/https://www.google.com/search?q=%22Now%2C+it%27s+time+for+your+treatment.+Relax%2C+make+yourself+comfortable.%22&oq=%22Now%2C+it%27s+time+for+your+treatment.+Relax%2C+make+yourself+comfortable.%22&aqs=chrome.0.69i59.536j0j9&sourceid=chrome&es_sm=0&ie=UTF-8",
        "google",
    )


def test_parse_query_google_lucius_vibullius_rufus_son_in_law_of_herodes_atticus_1642439444():
    verify_serp_parsing(
        "https://web.archive.org/web/20220117171044id_/https://www.google.com/search?as_eq=wikipedia&q=%22Lucius+Vibullius+Rufus%22+son-in-law+of+Herodes+Atticus",
        "google",
    )


def test_parse_query_google_la_pipa_1635023533():
    verify_serp_parsing(
        "https://web.archive.org/web/20211023211213id_/https://www.google.com/search?as_eq=wikipedia&q=%22LA+PIPA%22",
        "google",
    )


def test_parse_query_google_forced_labor_in_california_1635373615():
    verify_serp_parsing(
        "https://web.archive.org/web/20211027222655id_/https://www.google.com/search?as_eq=wikipedia&q=%22Forced+labor+in+California%22",
        "google",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_aliexpress_barefoot_accessories_1508385234():
    verify_serp_parsing(
        "https://web.archive.org/web/20171019035354id_/https://www.aliexpress.com/w/wholesale-barefoot-accessories.html?site=glo&g=y&SearchText=barefoot+accessories&pvId=200001034-361187&attrRel=or",
        "aliexpress",
    )


def test_parse_query_aliexpress_sterling_silver_gemstone_1384701071():
    verify_serp_parsing(
        "https://web.archive.org/web/20131117151111id_/http://www.aliexpress.com/wholesale?SearchText=sterling+silver+gemstone&initiative_id=RS_20131116203418",
        "aliexpress",
    )


def test_parse_query_aliexpress_women_messenger_bags_1397198302():
    verify_serp_parsing(
        "https://web.archive.org/web/20140411063822id_/http://www.aliexpress.com/w/wholesale-Women-Messenger-Bags.html?g=y&SearchText=Women%2BMessenger%2BBags&CatId=100002612&pvId=200000329-200004094&shipCountry=us&initiative_id=SB_20130813182932",
        "aliexpress",
    )


def test_parse_query_aliexpress_q5_phone_1374387535():
    verify_serp_parsing(
        "https://web.archive.org/web/20130721061855id_/http://www.aliexpress.com/wholesale/wholesale-q5-phone.html?SearchText=q5%2Bphone&CatId=0&SortType=total_weight_score_desc",
        "aliexpress",
    )


def test_parse_query_aliexpress_sterling_silver_pandora_charms_1457090417():
    verify_serp_parsing(
        "https://web.archive.org/web/20160304112017id_/http://www.aliexpress.com:80/w/wholesale-sterling-silver-pandora-charms.html?site=glo&shipCountry=us&g=y&SearchText=sterling+silver+pandora+charms&pvId=14-496&CatId=200160001",
        "aliexpress",
    )


def test_parse_query_aliexpress_children_school_bag_1388612710():
    verify_serp_parsing(
        "https://web.archive.org/web/20140101214510id_/http://www.aliexpress.com/w/wholesale-CHILDREN-SCHOOL-BAG.html?SearchText=CHILDREN%2BSCHOOL%2BBAG&CatId=380520&initiative_id=SB_20130923225029&SortType=total_weight_score_desc",
        "aliexpress",
    )


def test_parse_query_aliexpress_transmission_shaft_1435605057():
    verify_serp_parsing(
        "https://web.archive.org/web/20150629191057id_/http://www.aliexpress.com/w/wholesale-transmission-shaft.html?SearchText=transmission%2Bshaft&CatId=0&pvId=200001211-2277&shipCountry=us",
        "aliexpress",
    )


def test_parse_query_aliexpress_jewelry_sets_1390214039():
    verify_serp_parsing(
        "https://web.archive.org/web/20140120103359id_/http://www.aliexpress.com/w/wholesale-Jewelry-Sets.html?g=y&SearchText=Jewelry%2BSets&CatId=100006750&pvId=200000785-200003778%2C20252-1551&initiative_id=SB_20130731213711",
        "aliexpress",
    )


def test_parse_query_aliexpress_kids_wear_1293514157():
    verify_serp_parsing(
        "https://web.archive.org/web/20101228052917id_/http://www.aliexpress.com:80/wholesale?SearchText=kids+wear&CatId=0",
        "aliexpress",
    )


def test_parse_query_aliexpress_waterproof_winter_boots_women_1500418630():
    verify_serp_parsing(
        "https://web.archive.org/web/20170718225710id_/https://www.aliexpress.com/w/wholesale-waterproof-winter-boots-women.html?site=glo&g=y&SearchText=waterproof+winter+boots+women&pvId=200009123-1886&attrRel=or",
        "aliexpress",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_imdb_pulse_1283006912():
    verify_serp_parsing(
        "https://web.archive.org/web/20100828144832id_/http://www.imdb.com:80/find?s=all&q=Pulse",
        "imdb",
    )


def test_parse_query_imdb_0609265_s_nm_1329020836():
    verify_serp_parsing(
        "https://web.archive.org/web/20120212042716id_/http://www.imdb.com:80/find?q=0609265;s=nm",
        "imdb",
    )


def test_parse_query_imdb_angelina_jolie_1452700725():
    verify_serp_parsing(
        "https://web.archive.org/web/20160113155845id_/http://www.imdb.com:80/find?ref_=nv_sr_fn&q=Angelina+Jolie&s=all",
        "imdb",
    )


def test_parse_query_imdb_marcela_gomez_montoya_1614546944():
    verify_serp_parsing(
        "https://web.archive.org/web/20210228211544id_/http://www.imdb.com/find?q=Marcela%20G%C3%B3mez%20Montoya&exact=true",
        "imdb",
    )


def test_parse_query_imdb_murder_world_1268209692():
    verify_serp_parsing(
        "https://web.archive.org/web/20100310082812id_/http://www.imdb.com:80/find?s=all&q=murder+world",
        "imdb",
    )


def test_parse_query_imdb_gundula_rapsch_1628094679():
    verify_serp_parsing(
        "https://web.archive.org/web/20210804163119id_/https://www.imdb.com/find?q=Gundula+Rapsch&s=nm",
        "imdb",
    )


def test_parse_query_imdb_sam_claflin_1472223834():
    verify_serp_parsing(
        "https://web.archive.org/web/20160826150354id_/http://www.imdb.com:80/find?q=Sam%20Claflin&s=nm",
        "imdb",
    )


def test_parse_query_imdb_the_expanse_1521743964():
    verify_serp_parsing(
        "https://web.archive.org/web/20180322183924id_/http://www.imdb.com:80/find?s=all&q=the+expanse",
        "imdb",
    )


def test_parse_query_imdb_dogville_1187303706():
    verify_serp_parsing(
        "https://web.archive.org/web/20070816223506id_/http://www.imdb.com:80/find?s=all&q=dogville",
        "imdb",
    )


def test_parse_query_imdb_hunger_games_1518585690():
    verify_serp_parsing(
        "https://web.archive.org/web/20180214052130id_/http://www.imdb.com:80/find?q=hunger+games&s=tt&ref_=fn_tt",
        "imdb",
    )



from datetime import datetime, timezone
from itertools import chain
from os.path import getmtime
from pathlib import Path
from typing import Iterable, Iterator, NamedTuple
from urllib.parse import unquote
from uuid import uuid5

from click import echo
from elasticsearch_dsl.query import Term
from tqdm.auto import tqdm

from archive_query_log.config import Config
from archive_query_log.legacy.model import ArchivedUrl
from archive_query_log.legacy.urls.iterable import ArchivedUrls
from archive_query_log.namespaces import NAMESPACE_CAPTURE
from archive_query_log.orm import Capture, Archive, Provider, \
    InnerProvider, InnerArchive
from archive_query_log.utils.time import CET, UTC


class _ImportablePath(NamedTuple):
    path: Path
    archive: Archive
    provider: Provider
    domain: str
    url_path_prefix: str


def _iter_captures(
        config: Config,
        importable_path: _ImportablePath,
        last_modified: datetime,
        archived_urls: Iterable[ArchivedUrl],
        check_memento: bool = True,
) -> Iterator[Capture]:
    for archived_url in archived_urls:
        url = archived_url.url
        timestamp = datetime.fromtimestamp(
            archived_url.timestamp,
            tz=timezone.utc,
        )
        # Bug fix because the AQL-22 data is in CET, but the timestamps are
        # not marked as such.
        timestamp = timestamp.astimezone(CET)
        timestamp = timestamp.replace(tzinfo=UTC)

        memento_url = (
            f"{importable_path.archive.memento_api_url}/"
            f"{timestamp.astimezone(UTC).strftime('%Y%m%d%H%M%S')}/"
            f"{url}")
        if check_memento:
            response = config.http.session_no_retry.head(
                memento_url,
                allow_redirects=False,
            )
            if response.status_code != 200:
                continue

        capture_id_components = (
            importable_path.archive.cdx_api_url,
            url,
            timestamp.astimezone(UTC).strftime("%Y%m%d%H%M%S"),
        )
        capture_id = str(uuid5(
            NAMESPACE_CAPTURE,
            ":".join(capture_id_components),
        ))
        yield Capture(
            id=capture_id,
            last_modified=last_modified.replace(microsecond=0),
            archive=InnerArchive(
                id=importable_path.archive.id,
                cdx_api_url=importable_path.archive.cdx_api_url,
                memento_api_url=importable_path.archive.memento_api_url,
            ),
            provider=InnerProvider(
                id=importable_path.provider.id,
                domain=importable_path.domain,
                url_path_prefix=importable_path.url_path_prefix,
            ),
            url=url,
            timestamp=timestamp.astimezone(UTC),
            url_query_parser=InnerProvider(
                should_parse=True,
            ),
        )


def _import_captures_path(
        config: Config,
        importable_path: _ImportablePath,
        check_memento: bool = True,
) -> None:
    echo(f"Importing captures from {importable_path.path} to "
         f"archive {importable_path.archive.id} and "
         f"provider {importable_path.provider.id}.")

    json_paths = list(importable_path.path.glob("*.jsonl.gz"))
    oldest_modification_time = min(
        datetime.fromtimestamp(getmtime(path))
        for path in json_paths
    ).astimezone(UTC)
    echo(f"Found {len(json_paths)} JSONL files "
         f"(oldest from {oldest_modification_time.strftime('%c')}).")

    urls_iterators_list = [ArchivedUrls(path) for path in json_paths]
    urls_iterators: Iterable[ArchivedUrls] = urls_iterators_list
    if len(urls_iterators_list) > 50:
        # noinspection PyTypeChecker
        urls_iterators = tqdm(
            urls_iterators,
            desc="Get capture count",
            unit="file",
        )
    total_count = sum(len(urls) for urls in urls_iterators)
    echo(f"Found {total_count} captures.")

    archived_urls: Iterable[ArchivedUrl] = chain.from_iterable(
        urls_iterators_list)
    # noinspection PyTypeChecker
    archived_urls = tqdm(
        archived_urls,
        total=total_count,
        desc="Importing captures",
        unit="capture",
    )
    captures_iter = _iter_captures(
        config=config,
        importable_path=importable_path,
        last_modified=oldest_modification_time,
        archived_urls=archived_urls,
        check_memento=check_memento,
    )
    actions = (
        _create_action(capture, config)
        for capture in captures_iter
    )
    config.es.bulk(actions)


def _create_action(capture: Capture, config: Config) -> dict:
    capture.meta.index = config.es.index_captures
    return {
        **capture.to_dict(include_meta=True),
        "_op_type": "create",
    }


def import_captures(
        config: Config,
        data_dir_path: Path,
        check_memento: bool,
        search_provider: str | None,
        search_provider_index: int | None,
) -> None:
    echo(f"Importing AQL-22 captures from: {data_dir_path}")

    archive_response = (
        Archive.search(using=config.es.client, index=config.es.index_archives)
        .query(
            Term(cdx_api_url="https://web.archive.org/cdx/search/cdx")
        )
        .execute()
    )
    if archive_response.hits.total.value < 1:
        echo("No AQL-22 archive found. Add an archive with the "
             "CDX API URL 'https://web.archive.org/cdx/search/cdx' "
             "first.")
        exit(1)

    archive: Archive = archive_response.hits[0]
    echo(f"Importing captures for archive {archive.id}: {archive.name}")

    archived_urls_path = data_dir_path / "archived-urls"
    if not archived_urls_path.exists():
        echo("No captures found.")
        return

    search_provider_paths = sorted(archived_urls_path.glob("*"))
    if search_provider is not None:
        search_provider_paths = [
            path for path in search_provider_paths
            if path.name == search_provider
        ]
    elif search_provider_index is not None:
        search_provider_paths = [search_provider_paths[search_provider_index]]
    if len(search_provider_paths) == 0:
        echo("No captures found.")
        return
    prefix_paths_list: list[Path] = list(chain.from_iterable((
        search_provider_path.glob("*")
        for search_provider_path in search_provider_paths
    )))

    importable_paths = []
    prefix_paths: Iterable[Path] = prefix_paths_list
    # noinspection PyTypeChecker
    prefix_paths = tqdm(
        prefix_paths,
        desc="Checking URL prefixes",
        unit="prefix",
    )
    for prefix_path in prefix_paths:
        prefix = unquote(prefix_path.name)
        domain = prefix.split("/", maxsplit=1)[0]
        url_path_prefix = prefix.removeprefix(domain)

        provider_response = (
            Provider.search(using=config.es.client, index=config.es.index_providers)
            .query(
                Term(domains=domain) &
                Term(url_path_prefixes=url_path_prefix)
            )
            .execute()
        )
        if provider_response.hits.total.value >= 1:
            provider: Provider = provider_response.hits[0]
            importable_paths.append(_ImportablePath(
                path=prefix_path,
                archive=archive,
                provider=provider,
                domain=domain,
                url_path_prefix=url_path_prefix,
            ))

    for importable_path in importable_paths:
        _import_captures_path(
            config=config,
            importable_path=importable_path,
            check_memento=check_memento,
        )



from asyncio import run
from csv import reader
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from io import TextIOWrapper
from itertools import islice
from math import floor, log10
from pathlib import Path
from tempfile import gettempdir
from typing import Sized, Iterable, Any, Iterator, Mapping, Set, NamedTuple
from zipfile import ZipFile

from publicsuffixlist import PublicSuffixList
from ranx import Run, fuse
from requests import get, HTTPError
from requests.exceptions import ChunkedEncodingError
from tqdm.auto import tqdm

from archive_query_log.legacy.model import ArchivedUrl
from archive_query_log.legacy.download.raw import WebArchiveRawDownloader
from archive_query_log.legacy.util.http_session import backoff_session


@dataclass(frozen=True)
class AlexaTop1MArchivedUrls(Sized, Iterable[ArchivedUrl]):
    """
    Get all archived URLs of Alexa top-1M rankings.
    """
    output_path: Path
    cdx_api_url: str

    @cached_property
    def _params(self) -> Iterable[tuple[str, Any]]:
        return (
            ("url", "s3.amazonaws.com/alexa-static/top-1m.csv.zip"),
            ("fl", "timestamp,original"),
            ("filter", "mimetype:application/zip"),
            ("filter", "statuscode:200"),
        )

    @cached_property
    def _result_path(self) -> Path:
        return self.output_path

    @cached_property
    def _cache_path(self) -> Path:
        cache_path = Path(gettempdir()) / self.output_path.stem
        cache_path.mkdir(exist_ok=True)
        return cache_path

    @cached_property
    def num_pages(self) -> int:
        num_pages_response = get(
            self.cdx_api_url,
            params=[
                *self._params,
                ("showNumPages", True),
            ],
            # 10 minutes
            timeout=10 * 60  # nosec: B113
        )
        return int(num_pages_response.text)

    def _page_cache_path(self, page: int) -> Path:
        num_digits = floor(log10(self.num_pages)) + 1
        return self._cache_path / f"page_{page:{num_digits}}.jsonl"

    def _fetch_page(self, page: int) -> None:
        path = self._page_cache_path(page)
        if path.exists():
            # Page was already downloaded, skip it.
            if not path.is_file():
                raise RuntimeError(f"Path must be a file: {path}")
            return

        session = backoff_session()
        try:
            response = session.get(
                self.cdx_api_url,
                params=[
                    *self._params,
                    ("page", page),
                ],
                timeout=10 * 60  # 10 minutes, better safe than sorry ;)
            )
        except HTTPError:
            print(f"Failed to load page: {page}")
            return None
        except ChunkedEncodingError:
            print(f"Failed to read page contents: {page}")
            return None
        schema = ArchivedUrl.schema()
        with path.open("wt") as file:
            for line in response.text.splitlines(keepends=False):
                timestamp_string, url = line.split()
                timestamp = datetime.strptime(
                    timestamp_string,
                    "%Y%m%d%H%M%S"
                )
                archived_url = ArchivedUrl(url, int(timestamp.timestamp()))
                file.write(schema.dumps(archived_url))
                file.write("\n")

    def _fetch_pages(self) -> None:
        """
        Fetch queries from each individual page.
        """
        pages: Iterable[int] = range(self.num_pages)
        # noinspection PyTypeChecker
        pages = tqdm(
            pages,
            desc="Fetch urls",
            unit="page",
        )
        for page in pages:
            self._fetch_page(page)

    def _missing_pages(self) -> set[int]:
        """
        Find missing pages.
        Most often, the missing pages are caused by request timeouts.
        """
        missing_pages = set()
        for page in range(self.num_pages):
            path = self._page_cache_path(page)
            if not path.exists() or not path.is_file():
                missing_pages.add(page)
        return missing_pages

    def _merge_cached_pages(self) -> None:
        """
        Merge queries from all pages.
        """
        with self.output_path.open("wt", encoding="utf8") as file:
            pages: Iterable[int] = range(self.num_pages)
            # noinspection PyTypeChecker
            pages = tqdm(
                pages,
                desc="Merge urls",
                unit="page",
            )
            for page in pages:
                path = self._page_cache_path(page)
                with path.open("rt") as page_file:
                    lines = page_file
                    for line in lines:
                        file.write(line)

    def fetch(self) -> None:
        if self.output_path.exists():
            if not self.output_path.is_file():
                raise RuntimeError(f"Path must be a file: {self.output_path}")
            return
        print(f"Storing temporary files at: {self._cache_path}")
        self._fetch_pages()
        missing_pages = self._missing_pages()
        if len(missing_pages) > 0:
            raise RuntimeError(
                f"Pages missing: {missing_pages}\n"
                f"Consider retrying the download, as some requests "
                f"might only have timed out."
            )
        self._merge_cached_pages()
        for path in self._cache_path.iterdir():
            path.unlink()
        self._cache_path.rmdir()

    def __len__(self) -> int:
        self.fetch()
        with self.output_path.open("rt", encoding="utf8") as file:
            return sum(1 for _ in file)

    def __iter__(self) -> Iterator[ArchivedUrl]:
        self.fetch()
        schema = ArchivedUrl.schema()
        with self.output_path.open("rt", encoding="utf8") as file:
            for line in file:
                url = schema.loads(line, many=True)
                if isinstance(url, list):
                    raise ValueError(f"Expected one URL per line: {line}")
                yield url


def _iter_deduplicated(domains: Iterable[str]) -> Iterator[str]:
    public_suffix_list = PublicSuffixList()
    second_level_domains = set()
    for domain in domains:
        public_suffix = public_suffix_list.publicsuffix(domain)
        second_level_domain = public_suffix_list.subdomain(domain, 0)
        if second_level_domain is None:
            second_level_domain = public_suffix
        second_level_domain = second_level_domain.removesuffix(
            f".{public_suffix}"
        )
        if second_level_domain in second_level_domains:
            continue
        second_level_domains.add(second_level_domain)
        yield domain


class AlexaTop1MDomain(NamedTuple):
    rank: int
    domain: str
    public_suffix: str


@dataclass(frozen=True)
class AlexaTop1MFusedDomains(Sized, Iterable[AlexaTop1MDomain]):
    """
    Fuse the rop-1000 of all archived Alexa top-1M rankings.
    """
    data_directory_path: Path
    cdx_api_url: str
    fusion_method: str = "rrf"
    max_domains_per_ranking: int | None = 1000
    deduplicate_per_ranking: bool = True
    deduplicate_fused_ranking: bool = False

    @cached_property
    def _urls(self) -> Set[ArchivedUrl]:
        alexa_path = self.data_directory_path / \
                     "alexa-top-1m-archived-urls.jsonl"
        urls = AlexaTop1MArchivedUrls(
            output_path=alexa_path,
            cdx_api_url=self.cdx_api_url
        )
        return set(urls)

    @cached_property
    def _result_path(self) -> Path:
        name = f"alexa-top-1m-fused-domains-{self.fusion_method}" \
               f"-top-{self.max_domains_per_ranking}.csv"
        return self.data_directory_path / name

    @cached_property
    def _cache_path(self) -> Path:
        cache_path = Path(gettempdir()) / "alexa-top-1m-fused-domains"
        cache_path.mkdir(exist_ok=True)
        return cache_path

    def _fetch_rankings(self) -> Iterable[Path]:
        downloader = WebArchiveRawDownloader()
        paths: Mapping[ArchivedUrl, Path] = run(downloader.download(
            self._urls,
            self._cache_path,
            lambda url: f"{url.timestamp}.zip",
        ))
        if len(paths) < len(self._urls):
            raise RuntimeError("Some downloads were unsuccessful. Try again.")
        return paths.values()

    def _fuse_cached_rankings(self) -> None:
        runs: list[Run] = []
        num_runs = sum(1 for _ in self._cache_path.iterdir())
        paths: Iterable[Path] = self._cache_path.iterdir()
        # noinspection PyTypeChecker
        paths = tqdm(
            paths,
            total=num_runs,
            desc="Fuse rankings",
            unit="ranking",
        )
        for path in paths:
            with path.open("rb") as file:
                with ZipFile(file) as zip_file:
                    with zip_file.open("top-1m.csv", "r") as csv_file:
                        with TextIOWrapper(csv_file) as lines:
                            domains: Iterable[str] = (
                                line[1]
                                for line in reader(lines)
                            )
                            if self.deduplicate_per_ranking:
                                domains = _iter_deduplicated(domains)
                            if self.max_domains_per_ranking is not None:
                                domains = islice(
                                    domains,
                                    self.max_domains_per_ranking,
                                )
                            scores: dict[str, float] = {
                                domain: 1_000_000 - index
                                for index, domain in enumerate(domains)
                            }
                            alexa_run = Run({"_": scores})
                            runs.append(alexa_run)
        print(f"Fusing {len(runs)} rankings.")
        combined_run = fuse(
            runs=runs,
            norm="min-max",
            method=self.fusion_method,
        ).to_dict()
        items = sorted(
            combined_run["_"].items(),
            key=lambda item: item[1],
            reverse=True,
        )
        fused_domains: Iterable[str] = (domain for domain, _ in items)
        if self.deduplicate_fused_ranking and not self.deduplicate_per_ranking:
            fused_domains = _iter_deduplicated(fused_domains)
        public_suffix_list = PublicSuffixList(only_icann=True)
        with self._result_path.open("wt") as file:
            for index, domain in enumerate(fused_domains):
                public_suffix = public_suffix_list.publicsuffix(domain)
                file.write(f"{index + 1},{domain},{public_suffix}\n")

    def fetch(self) -> None:
        if self._result_path.exists():
            if not self._result_path.is_file():
                raise RuntimeError(f"Path must be a file: {self._result_path}")
            return
        print(f"Storing temporary files at: {self._cache_path}")
        self._fetch_rankings()
        self._fuse_cached_rankings()
        # for path in self._cache_path.iterdir():
        #     path.unlink()
        # self._cache_path.rmdir()

    def __len__(self) -> int:
        self.fetch()
        with self._result_path.open("rt") as file:
            return sum(1 for _ in file)

    def __iter__(self) -> Iterator[AlexaTop1MDomain]:
        self.fetch()
        with self._result_path.open("rt") as file:
            for line in file:
                index, domain, public_suffix = line.split(",")
                yield AlexaTop1MDomain(int(index), domain, public_suffix)



from re import compile as pattern, IGNORECASE
from typing import Sequence, Protocol, runtime_checkable, Any, Mapping, Union

from marshmallow.fields import Field

from archive_query_log.legacy.model import (
    ArchivedUrl, ArchivedSearchResultSnippet, ArchivedRawSerp)


@runtime_checkable
class QueryParser(Protocol):
    def parse(self, url: "ArchivedUrl") -> str | None:
        ...


@runtime_checkable
class PageParser(Protocol):
    def parse(self, url: "ArchivedUrl") -> int | None:
        ...


@runtime_checkable
class OffsetParser(Protocol):
    def parse(self, url: "ArchivedUrl") -> int | None:
        ...


@runtime_checkable
class InterpretedQueryParser(Protocol):
    def parse(self, raw_serp: "ArchivedRawSerp") -> str | None:
        ...


@runtime_checkable
class ResultsParser(Protocol):
    def parse(
            self,
            raw_serp: "ArchivedRawSerp",
    ) -> Sequence["ArchivedSearchResultSnippet"] | None:
        ...


class QueryParserField(Field):
    def _deserialize(
            self,
            value: Mapping[str, Any],
            attr: str | None,
            data: Mapping[str, Any] | None,
            **kwargs,
    ) -> QueryParser:
        parser_type = value["type"]
        if parser_type == "query_parameter":
            from archive_query_log.legacy.queries.parse import \
                QueryParameterQueryParser
            return QueryParameterQueryParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                parameter=value["parameter"],
            )
        elif parser_type == "fragment_parameter":
            from archive_query_log.legacy.queries.parse import \
                FragmentParameterQueryParser
            return FragmentParameterQueryParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                parameter=value["parameter"],
            )
        elif parser_type == "path_segment":
            from archive_query_log.legacy.queries.parse import \
                PathSegmentQueryParser
            return PathSegmentQueryParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                segment=value["segment"],
                remove_patterns=(
                    [
                        pattern(remove_pattern, IGNORECASE)
                        for remove_pattern in value["remove_patterns"]
                    ]
                    if "remove_patterns" in value
                    else []
                ),
                space_patterns=(
                    [
                        pattern(space_pattern, IGNORECASE)
                        for space_pattern in value["space_patterns"]
                    ]
                    if "space_patterns" in value
                    else []
                ),
            )
        elif parser_type == "fragment_segment":
            from archive_query_log.legacy.queries.parse import \
                FragmentSegmentQueryParser
            return FragmentSegmentQueryParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                segment=value["segment"],
                remove_patterns=(
                    [
                        pattern(remove_pattern, IGNORECASE)
                        for remove_pattern in value["remove_patterns"]
                    ]
                    if "remove_patterns" in value
                    else []
                ),
                space_patterns=(
                    [
                        pattern(space_pattern, IGNORECASE)
                        for space_pattern in value["space_patterns"]
                    ]
                    if "space_patterns" in value
                    else []
                ),
            )
        else:
            raise ValueError(f"Unknown parser type: {parser_type}")


class PageOffsetParserField(Field):
    def _deserialize(
            self,
            value: Mapping[str, Any],
            attr: str | None,
            data: Mapping[str, Any] | None,
            **kwargs,
    ) -> Union[PageParser | OffsetParser]:
        parser_type = value["type"]
        if parser_type == "query_parameter":
            from archive_query_log.legacy.queries.parse import \
                QueryParameterPageOffsetParser
            return QueryParameterPageOffsetParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                parameter=value["parameter"],
            )
        elif parser_type == "fragment_parameter":
            from archive_query_log.legacy.queries.parse import \
                FragmentParameterPageOffsetParser
            return FragmentParameterPageOffsetParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                parameter=value["parameter"],
            )
        elif parser_type == "path_segment":
            from archive_query_log.legacy.queries.parse import \
                PathSegmentPageOffsetParser
            return PathSegmentPageOffsetParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                segment=value["segment"],
                delimiter=value["delimiter"] if "delimiter" in value else "/",
                remove_patterns=(
                    [
                        pattern(remove_pattern, IGNORECASE)
                        for remove_pattern in value["remove_patterns"]
                    ]
                    if "remove_patterns" in value
                    else []
                ),
            )
        elif parser_type == "fragment_segment":
            from archive_query_log.legacy.queries.parse import \
                FragmentSegmentPageOffsetParser
            return FragmentSegmentPageOffsetParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                segment=value["segment"],
                delimiter=value["delimiter"] if "delimiter" in value else "/",
                remove_patterns=(
                    [
                        pattern(remove_pattern, IGNORECASE)
                        for remove_pattern in value["remove_patterns"]
                    ]
                    if "remove_patterns" in value
                    else []
                ),
            )
        else:
            raise ValueError(f"Unknown parser type: {parser_type}")


class InterpretedQueryParserField(Field):
    def _deserialize(
            self,
            value: Mapping[str, Any],
            attr: str | None,
            data: Mapping[str, Any] | None,
            **kwargs,
    ) -> InterpretedQueryParser:
        parser_type = value["type"]
        if parser_type == "html_selector":
            from archive_query_log.legacy.results.parse import \
                HtmlSelectorInterpretedQueryParser
            return HtmlSelectorInterpretedQueryParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                query_selector=value["query_selector"],
                query_attribute=(
                    value["query_attribute"]
                    if "query_attribute" in value
                    else "value"
                ),
                query_text=(
                    value["query_text"]
                    if "query_text" in value
                    else False
                ),
            )
        else:
            raise ValueError(f"Unknown parser type: {parser_type}")


class ResultsParserField(Field):
    def _deserialize(
            self,
            value: Mapping[str, Any],
            attr: str | None,
            data: Mapping[str, Any] | None,
            **kwargs,
    ) -> ResultsParser:
        parser_type = value["type"]
        if parser_type == "html_selector":
            from archive_query_log.legacy.results.parse import \
                HtmlSelectorResultsParser
            return HtmlSelectorResultsParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
                results_selector=value["results_selector"],
                url_selector=value["url_selector"],
                url_attribute=(
                    value["url_attribute"]
                    if "url_attribute" in value
                    else "href"
                ),
                title_selector=value["title_selector"],
                snippet_selector=(
                    value["snippet_selector"]
                    if "snippet_selector" in value
                    else None
                ),
            )
        elif parser_type == "chatnoir":
            from archive_query_log.legacy.results.chatnoir import \
                ChatNoirResultsParser
            return ChatNoirResultsParser(
                url_pattern=pattern(value["url_pattern"], IGNORECASE),
            )
        else:
            raise ValueError(f"Unknown parser type: {parser_type}")



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_ecosia_financial_risk_tolerance_quiz_1643759873():
    verify_serp_parsing(
        "https://web.archive.org/web/20220201235753id_/https://www.ecosia.org/search?q=financial+risk+tolerance+quiz",
        "ecosia",
    )


def test_parse_query_ecosia_free_file_sharing_services_1638926359():
    verify_serp_parsing(
        "https://web.archive.org/web/20211208011919id_/https://www.ecosia.org/search?q=free+file+sharing+services",
        "ecosia",
    )


def test_parse_query_ecosia_how_to_be_more_generous_1590871266():
    verify_serp_parsing(
        "https://web.archive.org/web/20200530204106id_/https://www.ecosia.org/search?q=how+to+be+more+generous&addon=chrome&addonversion=3.2.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_west_point_nebraska_1644663254():
    verify_serp_parsing(
        "https://web.archive.org/web/20220212105414id_/https://www.ecosia.org/search?q=west+point+nebraska",
        "ecosia",
    )


def test_parse_query_ecosia_hallam_f_c_1596174558():
    verify_serp_parsing(
        "https://web.archive.org/web/20200731054918id_/https://www.ecosia.org/search?q=Hallam+F.C.",
        "ecosia",
    )


def test_parse_query_ecosia_google_image_scraper_1601781718():
    verify_serp_parsing(
        "https://web.archive.org/web/20201004032158id_/https://www.ecosia.org/search?q=google+image+scraper",
        "ecosia",
    )


def test_parse_query_ecosia_form_1694_instructions_1635977208():
    verify_serp_parsing(
        "https://web.archive.org/web/20211103220648id_/https://www.ecosia.org/search?q=form+1694+instructions",
        "ecosia",
    )


def test_parse_query_ecosia_where_are_my_chrome_extensions_1635276242():
    verify_serp_parsing(
        "https://web.archive.org/web/20211026192402id_/https://www.ecosia.org/search?q=where+are+my+chrome+extensions",
        "ecosia",
    )


def test_parse_query_ecosia_gmail_sign_in_1637276050():
    verify_serp_parsing(
        "https://web.archive.org/web/20211118225410id_/https://www.ecosia.org/search?q=gmail+sign+in",
        "ecosia",
    )


def test_parse_query_ecosia_world_deutsch_regional_europa_deutschland_hessen_landkreise_rheingau_taunus_kreis_stadte_und_gemeind_1595135090():
    verify_serp_parsing(
        "https://web.archive.org/web/20200719050450id_/https://www.ecosia.org/search?q=World+Deutsch+Regional+Europa+Deutschland+Hessen+Landkreise+Rheingau-Taunus-Kreis+St%C3%A4dte+und+Gemeinden+R%C3%BCdesheim+am+Rhein",
        "ecosia",
    )


def test_parse_query_ecosia_how_much_doordash_driver_make_1654857157():
    verify_serp_parsing(
        "https://web.archive.org/web/20220610103237id_/https://www.ecosia.org/search?q=how+much+doordash+driver+make",
        "ecosia",
    )


def test_parse_query_ecosia_virtual_studio_code_download_for_windows_10_1642714491():
    verify_serp_parsing(
        "https://web.archive.org/web/20220120213451id_/https://www.ecosia.org/search?q=virtual+studio+code+download+for+windows+10",
        "ecosia",
    )


def test_parse_query_ecosia_h_and_m_clothing_sale_1635863133():
    verify_serp_parsing(
        "https://web.archive.org/web/20211102142533id_/https://www.ecosia.org/search?q=h+and+m+clothing+sale",
        "ecosia",
    )


def test_parse_query_ecosia_laughter_stories_1640836502():
    verify_serp_parsing(
        "https://web.archive.org/web/20211230035502id_/https://www.ecosia.org/search?q=laughter+stories&addon=chrome&addonversion=4.0.7&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_what_is_a_description_1655190914():
    verify_serp_parsing(
        "https://web.archive.org/web/20220614071514id_/https://www.ecosia.org/search?q=what+is+a+description",
        "ecosia",
    )


def test_parse_query_ecosia_macglauglin_1640487447():
    verify_serp_parsing(
        "https://web.archive.org/web/20211226025727id_/https://www.ecosia.org/search?q=macglauglin&addon=chrome&addonversion=4.0.7&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_8_ft_to_cm_1629639998():
    verify_serp_parsing(
        "https://web.archive.org/web/20210822134638id_/https://www.ecosia.org/search?q=8+ft+to+cm&addon=chrome&addonversion=3.4.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_how_to_check_if_a_value_is_null_in_c_1639813391():
    verify_serp_parsing(
        "https://web.archive.org/web/20211218074311id_/https://www.ecosia.org/search?tt=e8eb07a6&q=how+to+check+if+a+value+is+null+in+C%2B%2B&addon=brave",
        "ecosia",
    )


def test_parse_query_ecosia_we_will_rock_you_23yt_1587407460():
    verify_serp_parsing(
        "https://web.archive.org/web/20200420183100id_/https://www.ecosia.org/search?q=We+Will+Rock+You%2523yt",
        "ecosia",
    )


def test_parse_query_ecosia_health_conditions_and_diseases_neurological_disorders_meningitis_1666663445():
    verify_serp_parsing(
        "https://web.archive.org/web/20221025020405id_/https://www.ecosia.org/search?q=Health+Conditions+and+Diseases+Neurological+Disorders+Meningitis",
        "ecosia",
    )


def test_parse_query_ecosia_barnes_1636235793():
    verify_serp_parsing(
        "https://web.archive.org/web/20211106215633id_/https://www.ecosia.org/search?+Noble=&amp=&filters=sid%3A%22dd4a6430-b264-14e7-06d4-0c88211ae6fa%22&q=Barnes+",
        "ecosia",
    )


def test_parse_query_ecosia_saltpeter_beds_1643653808():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131183008id_/https://www.ecosia.org/search?q=saltpeter+beds&addon=chrome&addonversion=4.0.7&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_whiteman_afb_public_affairs_1636203076():
    verify_serp_parsing(
        "https://web.archive.org/web/20211106125116id_/https://www.ecosia.org/search?q=whiteman+afb+public+affairs",
        "ecosia",
    )


def test_parse_query_ecosia_can_you_show_smoking_on_tv_1624555764():
    verify_serp_parsing(
        "https://web.archive.org/web/20210624172924id_/https://www.ecosia.org/search?q=can+you+show+smoking+on+tv&addon=chrome&addonversion=3.4.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_what_does_god_do_to_tratorirs_1590883066():
    verify_serp_parsing(
        "https://web.archive.org/web/20200530235746id_/https://www.ecosia.org/search?q=what+does+God+do+to+tratorirs&addon=chrome&addonversion=3.2.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_windows_11_activator_reddit_1637172885():
    verify_serp_parsing(
        "https://web.archive.org/web/20211117181445id_/https://www.ecosia.org/search?q=windows+11+activator+reddit",
        "ecosia",
    )


def test_parse_query_ecosia_git_create_pull_request_1655702772():
    verify_serp_parsing(
        "https://web.archive.org/web/20220620052612id_/https://www.ecosia.org/search?q=git+create+pull+request",
        "ecosia",
    )


def test_parse_query_ecosia_hm_219c_1637494750():
    verify_serp_parsing(
        "https://web.archive.org/web/20211121113910id_/https://www.ecosia.org/search?q=hm+219c",
        "ecosia",
    )


def test_parse_query_ecosia_deaf_community_1626760345():
    verify_serp_parsing(
        "https://web.archive.org/web/20210720055225id_/https://www.ecosia.org/search?q=deaf+community&addon=chrome&addonversion=3.4.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_world_deutsch_regional_europa_deutschland_baden_wurttemberg_stadte_und_gemeinden_karlsruhe_1621530240():
    verify_serp_parsing(
        "https://web.archive.org/web/20210520170400id_/https://www.ecosia.org/search?q=World+Deutsch+Regional+Europa+Deutschland+Baden-W%C3%BCrttemberg+St%C3%A4dte+und+Gemeinden+Karlsruhe",
        "ecosia",
    )


def test_parse_query_ecosia_fudanshi_1655238275():
    verify_serp_parsing(
        "https://web.archive.org/web/20220614202435id_/https://www.ecosia.org/search?q=fudanshi",
        "ecosia",
    )


def test_parse_query_ecosia_why_can_t_you_order_international_nail_polish_online_india_23a_1586448930():
    verify_serp_parsing(
        "https://web.archive.org/web/20200409161530id_/https://www.ecosia.org/search?q=why+can%27t+you+order+international+nail+polish+online+india%2523a",
        "ecosia",
    )


def test_parse_query_ecosia_coding_freelancing_1634534706():
    verify_serp_parsing(
        "https://web.archive.org/web/20211018052506id_/https://www.ecosia.org/search?tt=e8eb07a6&q=coding+freelancing&addon=brave",
        "ecosia",
    )


def test_parse_query_ecosia_vasectomy_reversal_success_rate_23yt_1623229800():
    verify_serp_parsing(
        "https://web.archive.org/web/20210609091000id_/https://www.ecosia.org/search?q=vasectomy+reversal+success+rate%2523yt",
        "ecosia",
    )


def test_parse_query_ecosia_hisotry_of_tomato_in_india_23yt_1591696836():
    verify_serp_parsing(
        "https://web.archive.org/web/20200609100036id_/https://www.ecosia.org/search?q=hisotry+of+tomato+in+india%2523yt",
        "ecosia",
    )


def test_parse_query_ecosia_gnome_desktop_entry_1655209403():
    verify_serp_parsing(
        "https://web.archive.org/web/20220614122323id_/https://www.ecosia.org/search?q=gnome+desktop+entry",
        "ecosia",
    )


def test_parse_query_ecosia_game_changer_classic_1638948940():
    verify_serp_parsing(
        "https://web.archive.org/web/20211208073540id_/https://www.ecosia.org/search?q=game+changer+classic",
        "ecosia",
    )


def test_parse_query_ecosia_harry_styles_siblings_1634695267():
    verify_serp_parsing(
        "https://web.archive.org/web/20211020020107id_/https://www.ecosia.org/search?q=harry+styles+siblings",
        "ecosia",
    )


def test_parse_query_ecosia_what_s_the_shortest_war_ever_1635275630():
    verify_serp_parsing(
        "https://web.archive.org/web/20211026191350id_/https://www.ecosia.org/search?q=what%27s+the+shortest+war+ever",
        "ecosia",
    )


def test_parse_query_ecosia_sin_is_an_ugly_creature_1590737798():
    verify_serp_parsing(
        "https://web.archive.org/web/20200529073638id_/https://www.ecosia.org/search?q=sin+is+an+ugly+creature&addon=chrome&addonversion=3.2.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_high_functioning_depression_23w_1615448305():
    verify_serp_parsing(
        "https://web.archive.org/web/20210311073825id_/https://www.ecosia.org/search?q=high+functioning+depression%2523w",
        "ecosia",
    )


def test_parse_query_ecosia_correct_way_to_wash_hair_oily_scalp_1588676008():
    verify_serp_parsing(
        "https://web.archive.org/web/20200505105328id_/https://www.ecosia.org/search?q=correct+way+to+wash+hair+oily+scalp&addon=chrome&addonversion=3.1.2",
        "ecosia",
    )


def test_parse_query_ecosia_by_talos_this_cant_be_happening_1630519906():
    verify_serp_parsing(
        "https://web.archive.org/web/20210901181146id_/https://www.ecosia.org/search?q=%22by+talos+this+can%E2%80%99t+be+happening%22&addon=chrome&addonversion=3.4.0&method=topbar",
        "ecosia",
    )


def test_parse_query_ecosia_foot_to_centimeter_1612629786():
    verify_serp_parsing(
        "https://web.archive.org/web/20210206164306id_/https://www.ecosia.org/search?q=foot+to+centimeter",
        "ecosia",
    )


def test_parse_query_ecosia_percy_jackson_s_greek_gods_1634509308():
    verify_serp_parsing(
        "https://web.archive.org/web/20211017222148id_/https://www.ecosia.org/search?tt=e8eb07a6&q=percy+jackson%27s+greek+gods&addon=brave",
        "ecosia",
    )


def test_parse_query_ecosia_average_amount_of_photos_you_could_take_on_a_camera_in_1940_s_1572660962():
    verify_serp_parsing(
        "https://web.archive.org/web/20191102021602id_/https://www.ecosia.org/search?q=average+amount+of+photos+you+could+take+on+a+camera+in+1940%27s&addon=chrome&addonversion=2.1.0",
        "ecosia",
    )


def test_parse_query_ecosia_dall_e_1654952788():
    verify_serp_parsing(
        "https://web.archive.org/web/20220611130628id_/https://www.ecosia.org/search?tt=e8eb07a6&q=dall+e&addon=brave",
        "ecosia",
    )


def test_parse_query_ecosia_grady_white_ce_251_1636884788():
    verify_serp_parsing(
        "https://web.archive.org/web/20211114101308id_/https://www.ecosia.org/search?q=grady+white+ce+251",
        "ecosia",
    )


def test_parse_query_ecosia_windows_for_workgroups_3_11_iso_1644191359():
    verify_serp_parsing(
        "https://web.archive.org/web/20220206234919id_/https://www.ecosia.org/search?q=windows+for+workgroups+3.11+iso",
        "ecosia",
    )


def test_parse_query_ecosia_watch_tv_series_23g_1602851877():
    verify_serp_parsing(
        "https://web.archive.org/web/20201016123757id_/https://www.ecosia.org/search?q=watch+tv+series%2523g",
        "ecosia",
    )



from typing import Sequence

import pandas as pd
import yaml
from pandas import concat, DataFrame

from archive_query_log.legacy import DATA_DIRECTORY_PATH
from archive_query_log.legacy.cli.external import \
    load_services, load_domains, \
    service_domains, load_url_prefixes, \
    load_query_parsers, query_parser, load_page_offset_parsers, \
    page_offset_parser_series

services_file = DATA_DIRECTORY_PATH / "selected-services_overwrite.yaml"


def get_spreadsheet_data(
        first_service="google", last_service="hrblock"
) -> DataFrame:
    """
    Get parser information from the Google spreadsheet as a dataframe
    :param first_service:   First service to obtain data for (including)
    :param last_service:    Last service to obtain data for (including)
    :return:                Dataframe with parser information for each service
    """
    services = load_services()
    idx_first = services["name"].ne(first_service).idxmin()
    idx_last = services["name"].ne(last_service).idxmin()
    services = services.loc[idx_first:idx_last, :]  # type: ignore

    domains = load_domains()
    services["domains"] = [
        service_domains(domains, row)
        for _, row in services.iterrows()
    ]
    query_parsers = concat(
        [
            load_url_prefixes(),
            load_query_parsers()[["query_parser"]]
        ],
        axis="columns")
    query_parsers.dropna(inplace=True)
    services["query_parsers"] = [
        sorted((
            query_parser(row)
            for _, row in
            query_parsers[
                query_parsers["name"].str.fullmatch(service["name"])
            ].iterrows()
        ), key=lambda qp: str(qp["url_pattern"]))
        for _, service in services.iterrows()
    ]
    page_offset_parsers = concat(
        [
            load_url_prefixes(),
            load_page_offset_parsers()[["page_offset_parser"]]
        ],
        axis="columns")
    services["page_parsers"] = page_offset_parser_series(
        page_offset_parsers, services, count="pages")
    services["offset_parsers"] = page_offset_parser_series(
        page_offset_parsers, services, count="results")

    return services


def update_yaml_file(
        first_service="google", last_service="hrblock", overwrite=False
):
    """
    Update the local yaml file with the data from the Google spreadsheet
    :param first_service:   First service to update (including)
    :param last_service:    Last service to update (including)
    :param overwrite:       False: Only add parsers for service
                                that don't have one
                            True:   Overwrite the yaml entries
                                using the spreadsheet
    """
    services = get_spreadsheet_data(
        first_service=first_service, last_service=last_service)
    with open(services_file, "r", encoding="utf8") as stream:
        yaml_list = yaml.safe_load(stream)
    update_func = overwrite_parsers if overwrite else update_empty_parsers
    i = 0
    start_update = False
    while not start_update:
        elem = yaml_list[i]
        if elem["name"] == first_service:
            start_update = True
        else:
            i += 1

    while True:
        elem = yaml_list[i]
        name = elem["name"]
        update_func(service_elem=elem, services=services)
        i += 1
        if name == last_service:
            break

    with services_file.open("wt") as file:
        yaml.dump(yaml_list, stream=file, sort_keys=False)


def update_empty_parsers(service_elem: dict, services: DataFrame):
    offset_parsers = service_elem["offset_parsers"]
    page_parsers = service_elem["page_parsers"]
    query_parsers = service_elem["query_parsers"]

    # If no parsers are set, update the element using the services df
    if len(offset_parsers) + len(page_parsers) < 1:
        set_page_offset_parsers(service_elem=service_elem, services=services)
    if len(query_parsers) == 0:
        set_query_parsers(service_elem=service_elem, services=services)


def overwrite_parsers(service_elem: dict, services: DataFrame):
    set_page_offset_parsers(service_elem=service_elem, services=services)
    set_query_parsers(service_elem=service_elem, services=services)


def set_page_offset_parsers(service_elem: dict, services: DataFrame) -> None:
    name = service_elem["name"]
    row = services.loc[services["name"] == name, :]
    service_elem.update({"page_parsers": row["page_parsers"].values[0],
                         "offset_parsers": row["offset_parsers"].values[0]})


def set_query_parsers(service_elem: dict, services: DataFrame) -> None:
    name = service_elem["name"]
    row = services.loc[services["name"] == name, :]
    service_elem.update({"query_parsers": row["query_parsers"].values[0]})


def update_ranks(df: pd.DataFrame, yaml_list: Sequence[dict]):
    for i, elem in enumerate(yaml_list):
        name = elem["name"]
        rank_df = df.loc[df["service"] == name, "rank"]
        if len(rank_df) > 0:
            rank = int(rank_df.values[0])
        else:
            rank = 999999
        yaml_list[i]["alexa_rank"] = rank

    return yaml_list


def sort_by_rank(yaml_list: Sequence[dict]):
    return sorted(yaml_list, key=lambda d: d['alexa_rank'])



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_xvideos_drunk_1325431509():
    verify_serp_parsing(
        "https://web.archive.org/web/20120101152509id_/http://www.xvideos.com:80/?k=drunk",
        "xvideos",
    )


def test_parse_query_xvideos_escola_sao_luis_1657929903():
    verify_serp_parsing(
        "https://web.archive.org/web/20220716000503id_/https://www.xvideos.com/?k=escola+s%C3%A3o+lu%C3%ADs&sort=relevance&datef=all",
        "xvideos",
    )


def test_parse_query_xvideos_free_hardcore_porn_videos_japanese_video_1428541087():
    verify_serp_parsing(
        "https://web.archive.org/web/20150409005807id_/http://www.xvideos.com/?k=free-hardcore-porn-videos+japanese+video",
        "xvideos",
    )


def test_parse_query_xvideos_gay_whitezilla_go_to_www_lovesbo_ru_1643416941():
    verify_serp_parsing(
        "https://web.archive.org/web/20220129004221id_/https://www.xvideos.com/?k=gay+whitezilla+%E2%8F%A9+go+to+%E2%9D%A4%EF%B8%8F%EF%B8%8F+www.lovesbo.ru&sort=relevance&datef=week&typef=gay",
        "xvideos",
    )


def test_parse_query_xvideos_robber_1248193170():
    verify_serp_parsing(
        "https://web.archive.org/web/20090721161930id_/http://www.xvideos.com:80/?k=robber",
        "xvideos",
    )


def test_parse_query_xvideos_casey_cavert_go_to_www_lovesbo_ru_1642784433():
    verify_serp_parsing(
        "https://web.archive.org/web/20220121170033id_/https://www.xvideos.com/?k=casey+cavert+%E2%8F%A9+go+to+%E2%9D%A4%EF%B8%8F%EF%B8%8F+www.lovesbo.ru&sort=relevance&datef=week",
        "xvideos",
    )


def test_parse_query_xvideos_gay_peeing_porn_go_to_www_lovesbo_ru_1643357466():
    verify_serp_parsing(
        "https://web.archive.org/web/20220128081106id_/https://www.xvideos.com/?k=gay+peeing+porn+%E2%8F%A9+go+to+%E2%9D%A4%EF%B8%8F%EF%B8%8F+www.lovesbo.ru&sort=rating&typef=gay",
        "xvideos",
    )


def test_parse_query_xvideos_evander_marius_go_to_www_lovesbo_ru_1642716634():
    verify_serp_parsing(
        "https://web.archive.org/web/20220120221034id_/https://www.xvideos.com/?k=evander+marius+%E2%8F%A9+go+to+%E2%9D%A4%EF%B8%8F%EF%B8%8F+www.lovesbo.ru&sort=relevance&datef=3month",
        "xvideos",
    )


def test_parse_query_xvideos_claire_dames_1325320932():
    verify_serp_parsing(
        "https://web.archive.org/web/20111231084212id_/http://www.xvideos.com:80/?k=claire+dames",
        "xvideos",
    )


def test_parse_query_xvideos_solo_1341894522():
    verify_serp_parsing(
        "https://web.archive.org/web/20120710042842id_/http://www.xvideos.com:80/?k=solo&sort=relevance&durf=1-3min&datef=all",
        "xvideos",
    )
